# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

PRF.binary <- function(dataMatrix, groundtruth)
{
  dataMatrix  <- ifelse(dataMatrix  > 0, 1, 0)
  groundtruth <- ifelse(groundtruth > 0, 1, 0)
  tp = (groundtruth == dataMatrix) & (groundtruth == 1)
  fp = (groundtruth != dataMatrix) & (dataMatrix == 1)
  fn = (groundtruth != dataMatrix) & (dataMatrix == 0)
  pre = sum(tp) / sum(tp + fp)
  if(is.nan(pre))
    pre = 1
  rec = sum(tp) / sum(tp + fn)
  if(is.nan(rec))
    rec = 1
  fme = 2 * pre * rec / (pre + rec)
  if(is.nan(fme))
    fme = 0
  return(c(pre, rec, fme))
}

PRF <- function(dataMatrix, groundtruth)
{
  tp = ifelse(groundtruth > 0, dataMatrix / groundtruth, 0)
  fp = dataMatrix * ifelse(groundtruth>0,0,1)
  fn = ifelse(groundtruth > 0, groundtruth - dataMatrix / groundtruth, 0)
  pre = sum(tp) / sum(tp + fp)
  if(is.nan(pre))
    pre = 1
  rec = sum(tp) / sum(tp + fn)
  if(is.nan(rec))
    rec = 1
  fme = 2 * pre * rec / (pre + rec)
  if(is.nan(fme))
    fme = 0
  return(c(pre, rec, fme))
}

PRF.onsets <- function(dataMatrix, groundtruth, maxDist=10, offsets=FALSE, offsetTolerance=0.2, plotPRF=FALSE, stepSize=0.02, main="")   # 5 => +/- 100ms
{
#   shift = findOptimalShift(dataMatrix, groundtruth)
  
  # Thresholding (just in case)
  dataMatrix  <- ifelse(dataMatrix  > 0, 1, 0)
  groundtruth <- ifelse(groundtruth > 0, 1, 0)

  if(plotPRF)
    tp_matrix <- dataMatrix * 0

  detectedNotes <- matrix(0, 0, 2)
  trueNotes     <- matrix(0, 0, 2)
  tp <- fp <- fn <- dN <- tN <- 0
  for(ii in 1:nrow(dataMatrix))  # for each pitch
  {
    detectedNotes   = which(diff(c(0, dataMatrix[ii,], 0)) > 0)
    detectedOffsets = which(diff(c(0, dataMatrix[ii,], 0)) < 0)
    trueNotes       = which(diff(c(0, groundtruth[ii,], 0)) > 0)
    if(offsets)
      trueOffsets     = which(diff(c(0, groundtruth[ii,], 0)) < 0)
    dN <- dN + length(detectedNotes)
    tN <- tN + length(trueNotes)
    distances <- outer(trueNotes, detectedNotes, "-")  # distance matrix
    distances <- abs(distances)
    distances[distances > maxDist] <- +Inf
    local_tp <- 0
    while(!is.infinite(min(distances)))
    {
      minElement <- which.min(distances)
      minCol <- ceiling(minElement / nrow(distances))
      minRow <- minElement - (minCol - 1) * nrow(distances)

      if(offsets)
      {
        detectedLength = detectedOffsets[minCol] - detectedNotes[minCol]
        trueLength     = trueOffsets[minRow] - trueNotes[minRow]
      }

      distances[minRow,] <- +Inf  # we cannot match the same true note again
      distances[,minCol] <- +Inf  # we cannot match the same detected note again

      if(offsets)
        if((detectedLength/trueLength) < (1.0 - offsetTolerance) || (detectedLength/trueLength) > (1.0 + offsetTolerance))
          next

      local_tp <- local_tp + 1

      if(plotPRF)
        tp_matrix[ii, detectedNotes[minCol]:(detectedOffsets[minCol]-1)] <- 1
    }
    fp <- fp + length(detectedNotes) - local_tp
    fn <- fn + length(trueNotes)     - local_tp
    tp <- tp + local_tp
  }
  if(plotPRF)
  {
    fp_matrix <- dataMatrix * 0
    fp_matrix[dataMatrix > tp_matrix] <- 1
    fn_matrix <- dataMatrix * 0
    fn_matrix[trueNotes > tp_matrix] <- 1
    image(x=seq(0, by=stepSize, length.out=ncol(dataMatrix)), y=1:nrow(dataMatrix), z=t(tp_matrix + fp_matrix*2 + fn_matrix*3), zlim=c(0,3), col=c("white", "black", "red", "green"), xlab="Time [s]", ylab="Piano note number", main=main)
    text(stepSize*5, 85, paste("Detected notes : ", dN, sep=""), pos=4, col="black")
    text(stepSize*5, 81, paste("Real notes : ", tN, sep=""), pos=4, col="black")
    text(stepSize*5, 77, paste("True positives : ", tp, sep=""), pos=4, col="black")
    text(stepSize*5, 73, paste("False positives : ", fp, sep=""), pos=4, col="red")
    text(stepSize*5, 69, paste("False negatives : ", fn, sep=""), pos=4, col="red")
    for(ii in 1:nrow(dataMatrix))
    {
      xpoints <- stepSize * (which(diff(c(0, groundtruth[ii,], 0)) != 0) - 1)
      if(length(xpoints > 1))
      {
        for(pointnr in 1:(length(xpoints)/2))
          lines(x=xpoints[((pointnr-1)*2+1):(pointnr*2)], y=c(ii, ii), col="orange", lwd=2)
      }
    }
  }

  pre = tp / (tp + fp)
  if(is.nan(pre))
    pre = 1
  rec = tp / (tp + fn)
  if(is.nan(rec))
    rec = 1
  fme = 2 * pre * rec / (pre + rec)
  if(is.nan(fme))
    fme = 0
  return(c(pre, rec, fme))
}
