# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library("timeSeries")

source("Auxiliary/fancyOutput.R")

# ################################################################################################################
# averageLikelihoodFunction -- general function for calculating average cross-entropy on an entire dataset
#                              for a given model structure (H, HC, V, P, VP, HVP or HCVP)
#
averageLikelihoodFunction = function(fudgeFactors, dataset=mutopiaFiles.testing, models="HCVP", bins="all", interpolation="loglinear", silent=TRUE)
{
  cat(" -------------------------------------\n")
#   cat(" - Fudge factors =", exp(fudgeFactors), "\n")
#   modellikelihood = mean(calculateCrossEntropies(exp(fudgeFactors), dataset=dataset, models=models, bins=bins, silent=silent), na.rm=TRUE)
  cat(" - Fudge factors =", fudgeFactors, "\n")
  modellikelihood = mean(calculateCrossEntropies(fudgeFactors, dataset=dataset, models=models, bins=bins, interpolation=interpolation, silent=silent), na.rm=TRUE)
  cat(" - Average likelihood =", modellikelihood, "\n")

  tempFileName = paste("Temp/fudgeFactorsTrainingTemp.", models, ".loglinear.Rdata", sep="")
  if(file.exists(tempFileName))
    load(tempFileName)
  else
    tempResults = list()
  tempResults[[length(tempResults)+1]] = list(fudgeFactors=fudgeFactors, value=modellikelihood)
  save(tempResults, file=tempFileName)

  return(modellikelihood)
}

# bins = c("all", "onsets", "offsets", "changes", "silence", "notes")
calculateCrossEntropies = function(fudgeFactors, dataset=mutopiaFiles.testing, models="HCDVPN", bins="all", interpolation="loglinear", parallelize=TRUE, silent=FALSE)
{
  source("Training/probabilityDefinitions.R")
  uniformChordPty = 1/24
  uninformedPty = 2^-88

  harmonyModelUsed   = length(grep("H", models)) > 0
  chordModelUsed     = length(grep("C", models)) > 0
  durationModelUsed  = length(grep("D", models)) > 0
  voiceModelUsed     = length(grep("V", models)) > 0
  polyphonyModelUsed = length(grep("P", models)) > 0
  neighbourModelUsed = length(grep("N", models)) > 0

  if(!is.null(interpolation))
  {
    interpolationCoefficients = c(harmonyModelUsed, durationModelUsed, voiceModelUsed, polyphonyModelUsed, neighbourModelUsed)
    modelCount = length(which(interpolationCoefficients))
    interpolationCoefficients[which(interpolationCoefficients)] = 1:modelCount
    interpolationCoefficients[interpolationCoefficients == 0] = NA

    if(interpolation == "loglinear")
    {
      harmonyPties.fudged = harmonyPties^fudgeFactors[interpolationCoefficients[1]]
      durationPties.fudged = durationPties^fudgeFactors[interpolationCoefficients[2]]
      voicePties.fudged = simpleVoiceProbs^fudgeFactors[interpolationCoefficients[3]]
      polyphonyPties.fudged = polyphonyPties^fudgeFactors[interpolationCoefficients[4]]
      neighbourPties.fudged = neighbourPties^fudgeFactors[interpolationCoefficients[5]]
    }
    else if(interpolation == "linear")
    {
      fudgeFactors = fudgeFactors / sum(fudgeFactors)  # so that we don't need to normalize the resulting note prior
      harmonyPties.fudged = harmonyPties * fudgeFactors[interpolationCoefficients[1]]
      durationPties.fudged = durationPties * fudgeFactors[interpolationCoefficients[2]]
      voicePties.fudged = simpleVoiceProbs * fudgeFactors[interpolationCoefficients[3]]
      polyphonyPties.fudged = polyphonyPties * fudgeFactors[interpolationCoefficients[4]]
      neighbourPties.fudged = neighbourPties * fudgeFactors[interpolationCoefficients[5]]
    }
    else
      stop("Unknown interpolation method!")
  }

  calculateForSingleFile = function(filename, silent=FALSE)
  {
    cat("@ ")
    timeSpent = system.time({
      load(paste("Datasets/Mutopia/Processed/", filename, sep=""))
      dataSize = ncol(dataMatrix)

      # Select the bins to intergate over when calculating para-cross-entropy
      if(bins != "all")
      {
        onsetMatrix = matrix(FALSE, nrow=nrow(dataMatrix), ncol=ncol(dataMatrix))
        if(bins == "onsets")
        {
          for(rowNr in 1:88)
            onsetMatrix[rowNr,] = diff(c(0, dataMatrix[rowNr,])) == 1
        }
        else if(bins == "offsets")
        {
          for(rowNr in 1:88)
            onsetMatrix[rowNr,] = diff(c(0, dataMatrix[rowNr,])) == -1
        }
        else if(bins == "changes")
        {
          for(rowNr in 1:88)
            onsetMatrix[rowNr,] = diff(c(0, dataMatrix[rowNr,])) != 0
        }
        else if(bins == "silence")
          onsetMatrix = (dataMatrix == 0)
        else if(bins == "notes")
          onsetMatrix = (dataMatrix == 1)
        else
          stop("bins = ", bins, " unimplemented")
      }

      if(models == "Bernoulli")
      {
        noteProbabilities = t(ifelse(dataMatrix > 0, priorOnProbability, priorOffProbability))
      }
      else if(models == "PBernoulli")
      {
        noteProbabilities = t(ifelse(dataMatrix > 0, matrix(rep(priorOnProbabilities, dataSize), nrow=88), matrix(rep(priorOffProbabilities, dataSize), nrow=88)))
      }
      else
      {
        if(chordModelUsed)
          transitionMatrix = chordPty
        else
          transitionMatrix = uniformChordPty * matrix(1, 24, 24)
        onsetCounter = 0

        if(harmonyModelUsed)
          noteProbabilities = array(NA,dim=c(dataSize, 88, 24))
        else
          noteProbabilities = matrix(NA, dataSize, 88)
        for(chordNr in 1:ifelse(harmonyModelUsed, 24, 1))
        {
          if(harmonyModelUsed)
            unnormalizedProbs.H = t(harmonyPties.fudged[, chordNr, ])
          else
            unnormalizedProbs.H = matrix(1, 88, 2)

          for(frameNr in 1:dataSize)
          {
            unnormalizedProbs = unnormalizedProbs.H
            if(voiceModelUsed)
            {
              distances = rep(NA, 88)
              activeNotes = which(dataMatrix[, frameNr-1] > 0)
              if(length(activeNotes) > 0)
              {
                for(rowNr in 1:88)
                  distances[rowNr] = min(abs(rowNr - activeNotes)) + 1
              }
              else
                distances = rep(89, 88)

              if(interpolation=="loglinear")
                unnormalizedProbs = unnormalizedProbs * voicePties.fudged[distances, ]
              else if(interpolation=="linear")
                unnormalizedProbs = unnormalizedProbs + voicePties.fudged[distances, ]
            }
            if(polyphonyModelUsed)
            {
              notesBelow = (c(rev(cumsum(rev(dataMatrix[, frameNr]))), 0) + 1)[2:89]
              if(interpolation=="loglinear")
                unnormalizedProbs = unnormalizedProbs * polyphonyPties.fudged[notesBelow, ]
              else if(interpolation=="linear")
                unnormalizedProbs = unnormalizedProbs + polyphonyPties.fudged[notesBelow, ]
            }
            if(durationModelUsed)
            {
              if(interpolation=="loglinear")
              {
                if(frameNr == 1)
                  unnormalizedProbs = unnormalizedProbs * t(durationPties.fudged[,,1])
                else
                  unnormalizedProbs = unnormalizedProbs * cbind(durationPties.fudged[cbind(1,1:88,dataMatrix[,frameNr-1]+1)], durationPties.fudged[cbind(2,1:88,dataMatrix[,frameNr-1]+1)])
              }
              else if(interpolation=="linear")
              {
                if(frameNr == 1)
                  unnormalizedProbs = unnormalizedProbs + t(durationPties.fudged[,,1])
                else
                  unnormalizedProbs = unnormalizedProbs + cbind(durationPties.fudged[cbind(1,1:88,dataMatrix[,frameNr-1]+1)], durationPties.fudged[cbind(2,1:88,dataMatrix[,frameNr-1]+1)])
              }
            }
            if(neighbourModelUsed)
            {
              notesBelow = c(0, dataMatrix[1:87,frameNr]) + c(0, 0, dataMatrix[1:86,frameNr])*2 + 1
              if(interpolation=="loglinear")
                unnormalizedProbs = unnormalizedProbs * t(neighbourPties.fudged[,notesBelow])
              else if(interpolation=="linear")
                unnormalizedProbs = unnormalizedProbs + t(neighbourPties.fudged[,notesBelow])
            }

            unnormalizedProbs = unnormalizedProbs / rowSums(unnormalizedProbs) # normalization (unnecessary for linear interpolation, but does not take much time anyway)
            if(harmonyModelUsed)
              noteProbabilities[frameNr, , chordNr] = unnormalizedProbs[cbind(1:88, dataMatrix[, frameNr] + 1)]
            else
              noteProbabilities[frameNr,] = unnormalizedProbs[cbind(1:88, dataMatrix[, frameNr] + 1)]
          }
        }
      }
      
      if(harmonyModelUsed)  # we need to calculate the entropy with the forward/backward algorithm
      {
        forwardProbabilities = matrix(NA, 24, dataSize)
        forwardVectorNorms   = rep(NA, dataSize)
        forwardProbabilities[,1] = colProds(noteProbabilities[1,,]) * uniformChordPty
        forwardVectorNorms[1] = sum(forwardProbabilities[,1])
        forwardProbabilities[,1] = forwardProbabilities[,1] / forwardVectorNorms[1]  # normalization
        for(timeFrame in 2:dataSize)
        {
          forwardProbabilities[,timeFrame] = colProds(noteProbabilities[timeFrame,,]) * (forwardProbabilities[,timeFrame-1] %*% transitionMatrix)
          forwardVectorNorms[timeFrame] = sum(forwardProbabilities[,timeFrame])
          forwardProbabilities[,timeFrame] = forwardProbabilities[,timeFrame] / forwardVectorNorms[timeFrame]  # normalization
        }

        if(bins == "all")
        {
          modellikelihood = -sum(log2(forwardVectorNorms)) / prod(dim(dataMatrix))
        }
        else
        {
          entropyVectorNorms = rep(NA, dataSize)
          entropyVector      = colProds(noteProbabilities[1,onsetMatrix[,1],]) * uniformChordPty
          entropyVectorNorms[1] = sum(entropyVector)
          entropyVector      = entropyVector / forwardVectorNorms[1]  # normalization
          for(timeFrame in 2:dataSize)
          {
            nonoffsetForwardVector = colProds(noteProbabilities[timeFrame,!onsetMatrix[,timeFrame],]) * (forwardProbabilities[,timeFrame-1] %*% transitionMatrix) # P(C_t = i | N_{1:t} exluding current offsets)
            nonoffsetForwardVector = nonoffsetForwardVector / sum(nonoffsetForwardVector)
            entropyVector = colProds(noteProbabilities[timeFrame,onsetMatrix[,timeFrame],]) * nonoffsetForwardVector
            entropyVectorNorms[timeFrame] = sum(entropyVector)
            entropyVector = entropyVector / entropyVectorNorms[timeFrame]  # normalization
          }
          modellikelihood = -sum(log2(entropyVectorNorms[colSums(onsetMatrix) > 1])) / length(which(onsetMatrix))
        }
      }
      else
      {
#         if(bins == "onsets") recover()
        if(bins == "all")
          modellikelihood = sum(-log2(noteProbabilities)) / prod(dim(dataMatrix))
        else
          modellikelihood = sum(-log2(t(noteProbabilities)[as.vector(onsetMatrix)])) / length(which(onsetMatrix))

      }
    })[1]
    cat(modellikelihood, " ")

    return(list(modellikelihood=modellikelihood, timing=timeSpent))
  }

  if(!parallelize)
  {
    modellikelihoods = rep(NA, length(dataset))
    fileNr = 1
    timings = c()
    for(filename in dataset)
    {
      cat("\t· [", fileNr, "/", length(dataset), "] ", filename, "... ", sep="")
      singleResults = calculateForSingleFile(filename)
      modellikelihoods[fileNr] = singleResults$modellikelihood
      timings = c(timings, singleResults$timing)
      cat(singleResults$timing, " s (mean = ", mean(timings), " s)\n", sep="")
      fileNr = fileNr + 1
    }
  }
  else
  {
    library("multicore")
    if(!silent) cat("\t· Running calculations on multiple cores\n")
    allResults = simplify2array(mclapply(dataset, function(x) calculateForSingleFile(x, silent=FALSE)))
    modellikelihoods = unlist(allResults[1,])
  }

  return(modellikelihoods)
}

