# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

dataset = unlist(read.table("mutopiaFiles.txt", stringsAsFactors=FALSE), use.names=FALSE)

frameSize = 1/6  # in beats
ticksPerQuarterNote = 384 # it's the same for all Mutopia files

calculateFramePos = function(fileNr)
{
  filename = dataset[fileNr]
  cat("[", fileNr, "/", length(dataset), "] ", filename, "\n", sep="")

#   system(paste("listmidievents MIDI/", filename, " | grep TEMPO > tempfile.txt", sep=""))
  system(paste("~/bin/listmidievents ~/Dropbox/Artykuly/IEEE/Experiments.reimplemented/Datasets/Mutopia/MIDI/", filename, " | grep TEMPO > tempfile.txt", sep=""))
  tempos = data.frame(list("TEMPO", 0, 120, NA))
  try({
    tempos = read.table("tempfile.txt")              # tempos in beats per minute
  }, silent=TRUE)
  tempos[,2] = tempos[,2] / ticksPerQuarterNote    # times in quarter notes

#   system(paste("listmidievents MIDI/", filename, " | grep METER > tempfile.txt", sep=""))
  system(paste("~/bin/listmidievents ~/Dropbox/Artykuly/IEEE/Experiments.reimplemented/Datasets/Mutopia/MIDI/", filename, " | grep METER > tempfile.txt", sep=""))
  meters = data.frame(list("METER", 0, 4, 1))
  try({
    meters = read.table("tempfile.txt")
  }, silent=TRUE)
  meters[,2] = meters[,2] / ticksPerQuarterNote    # times in quarter notes
  meters[,4] = 2^(2-meters[,4])                    # the beat value in quarter notes (2=half note, 1=quarter note, 0.5=eighth, etc.)

  # We need to process the meter and tempo information to know the event times in beats
  colnames(meters) = colnames(tempos)
  timeEvents = rbind(tempos, meters)
  timeEvents = timeEvents[sort(timeEvents[,2], index.return=TRUE)$ix,]  # sort by the tick-time
  if(!any(timeEvents[timeEvents[,1] == "METER", 2] == 0))  # no meter specified at the beginning
    timeEvents = rbind(list("METER", 0, 4, 1), timeEvents)     # default is 4/4 time
  if(!any(timeEvents[timeEvents[,1] == "TEMPO", 2] == 0))  # no meter specified at the beginning
    timeEvents = rbind(list("TEMPO", 0, 120, NA), timeEvents)  # default is 120 quarter notes per minute
  eventCount = nrow(timeEvents)
  if(eventCount > 2)   # tempo/meter changes within the piece
  {
    currentBeatValue = timeEvents[which(timeEvents[,1] == "METER")[1], 4]
    timeEvents[, 2] = timeEvents[, 2] / currentBeatValue
    if(eventCount > 3)
      for(eventNr in 3:(nrow(timeEvents)-1))
      {
        if(timeEvents[eventNr, 1] == "METER")
        {
          newBeatValue = timeEvents[eventNr, 4]
          timeEvents[(eventNr+1):eventCount, 2] = timeEvents[eventNr, 2] + (timeEvents[(eventNr+1):eventCount, 2] - timeEvents[eventNr, 2]) * currentBeatValue / newBeatValue
          currentBeatValue = newBeatValue
        }
      }
  }

  tempos = data.frame(timeEvents[timeEvents[,1] == "TEMPO",])
  meters = data.frame(timeEvents[timeEvents[,1] == "METER",])

  processedFilename = gsub("/", "_", filename)
  processedFilename = gsub("\\.mid", "", processedFilename)
  load(paste("Processed/", processedFilename, ".txt.Rdata", sep=""))
  dataLength = ncol(dataMatrix)
  initialRest = which(diff(c(0, which(colSums(dataMatrix) == 0))) != 1)[1] - 1
                # R magic to count how many columns at the beginning of dataMatrix are empty
                # this is important because Timidity++ removes the initial silence when 
                # synthesizing MIDI files. Not removing the silence here causes all this script
                # to make tempo/meter changes in wrong places, effectively making the resulting
                # NMF results useless as ground truth, even if it is shiftedl later.
  if(is.na(initialRest))  # happens if there were no empty columns at all
    initialRest = 0

#   oldTempo = 0
#   tempoChanges = matrix(NA, 0, 3)
  framePositions = seq(from=initialRest+frameSize, by=frameSize, length.out=dataLength)  # frame positions in beats
  for(frameNr in seq_along(framePositions))
  {
    currentTempo     = tempos[rev(which(tempos[,2] < framePositions[frameNr]))[1], 3]  # in quarter notes per minute
    currentBeatValue = meters[rev(which(meters[,2] < framePositions[frameNr]))[1], 4]
    currentTempo     = (currentTempo / currentBeatValue) / 60   # in beats per second
    frameLength      = frameSize / currentTempo  # in seconds
    if(frameNr > 1)
      framePositions[frameNr] = framePositions[frameNr-1] + frameLength
    else
      framePositions[frameNr] = frameLength
#     if(oldTempo != currentTempo)
#     {
#       tempoChanges = rbind(tempoChanges, c(frameNr, framePositions[frameNr], currentTempo))
#       oldTempo = currentTempo
#     }
  }

#   plot(y=seq_along(framePositions)/6, x=framePositions, type="s", xlab="Time [s]", ylab="Beat", axes=F)
#   axis(2); box(); axis(1, at=seq(from=0, to=max(framePositions), by=15), labels=round(seq(from=0, to=max(framePositions), by=15), 2), las=2)
#   abline(h=tempoChanges[,1]/6, col="red")
#   text(x=rep(0, nrow(tempoChanges)), y=tempoChanges[,1]/6, round(tempoChanges[,3],2), col="red", pos=1)
#   abline(v=tempoChanges[,2], col="dark green")
#   text(y=rep(50, nrow(tempoChanges)), x=tempoChanges[,2], paste(floor(tempoChanges[,2] / 60), round(tempoChanges[,2] - floor(tempoChanges[,2] / 60) * 60, 0), sep=":"), col="dark green", pos=4)
#   recover()

  write.table(framePositions, paste("FramePos/", processedFilename, ".csv", sep=""), row.names=FALSE, col.names=FALSE)
  return(TRUE)
}

# library("multicore")  # we CANNOT use multicore because the temp files would mix up
results = sapply(1:length(dataset), calculateFramePos)

# calculateFramePos("SchubertF/SchubertF-D881B-Fischerweise.mid")
