# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

filename = commandArgs(T)[1]

mididata = read.table(paste("Events/", filename, ".txt", sep=""), header=F, stringsAsFactors=FALSE)

if(nrow(mididata) < 2)
{
  cat("This data file is too short!\n")
} else
{
  if(any(mididata[!is.na(mididata[,4]),4] == 0))
    mididata[which(mididata[,4] == 0), 1] = "NOTEOFF"

  ticksPerQuarterNote = mididata[mididata[,1] == "TPQ", 3]   # it's 384 for all the quantized MIDI files
  cat("Ticks per quarter note:", ticksPerQuarterNote, "\n")

  quarterNotesPerBeat = 1
  ticksPerBeat = ticksPerQuarterNote * quarterNotesPerBeat

  mididata = mididata[sort(mididata[,2], index.return=TRUE)$ix,]  # sort midi events by their time
  for(eventnr in 1:(nrow(mididata)-1))
  {
    if((mididata[eventnr, 1] == "METER") && (mididata[eventnr+1, 2] == mididata[eventnr, 2]))
    {
      tempVar = mididata[eventnr+1,]
      mididata[eventnr+1,] = mididata[eventnr,]
      mididata[eventnr,]   = tempVar
    }
  }

  deltatimes = c(0, diff(mididata[,2]))
  for(eventnr in 1:nrow(mididata))
  {
    if(eventnr == 1)
      deltatimes[eventnr] = deltatimes[eventnr] / ticksPerBeat
    else
      deltatimes[eventnr] = deltatimes[eventnr-1] + deltatimes[eventnr] / ticksPerBeat
    if(mididata[eventnr, 1] == "METER")
    {
      quarterNotesPerBeat = 2^(2 - mididata[eventnr, 3])
      cat("Time signature: ", mididata[eventnr,2], "/", 2^(mididata[eventnr,3]), "\n", sep="")
      ticksPerBeat = ticksPerQuarterNote * quarterNotesPerBeat
    }
  }
  mididata[,2] = deltatimes

  notelist = c()
  onsets  = which(mididata[, 1] == "NOTEON")
  offsets = which(mididata[, 1] == "NOTEOFF")
  for(onset in onsets)
  {
    note = rep(NA, 3)
    note[1] = mididata[onset,3]  # pitch
    note[2] = mididata[onset,2]  # onset position
    offsetpos = mididata[offsets,2][mididata[offsets,3] == note[1]]
    note[3] = offsetpos[offsetpos > note[2]][which.min(abs(offsetpos[offsetpos > note[2]] - note[2]))] # find note offset
    notelist = rbind(notelist, note)  # add note to the list
  }
  colnames(notelist) = c("pitch", "start", "end")
  notelist = notelist[sort(notelist[,2], index.return=TRUE)$ix,]

  write.table(notelist, file=paste("Notes.beat/", filename, ".txt", sep=""), row.names=FALSE)
  cat("Found", nrow(notelist), "notes\n")
}
