# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library("sound")

source("fileList.R")

timeBasis = 2   # don't change
frameSize = 1/6
framesPerBeat = 1/frameSize

MIR.Convert.FrequencyFromMIDInote <- function(note) return(440 * 2 ^ (note / 12))
MIR.Convert.FrequencyFromPianokey <- function(pianokey) MIR.Convert.FrequencyFromMIDInote(pianokey - 49)

createNMFbasis = function(kernelSize, peakslope=3, decay=1)
{
  PianokeyFromFrequency <- function(x) 12 * log(x/440., 2) + 49
  harmonicBasis <- matrix(0, kernelSize, 88)
  for(basisVectorNr in 1:ncol(harmonicBasis))
  {
    baseFrequency <- MIR.Convert.FrequencyFromPianokey(basisVectorNr)
    for(harmonic in 1:100)
    {
      peakPosition  <- (PianokeyFromFrequency(baseFrequency * harmonic) - 1) * 3 + 1
      if(peakPosition < nrow(harmonicBasis) - 1)
      {
        harmonicBasis[peakPosition, basisVectorNr] <- harmonic^(-decay)
        if((peakPosition > 1) && (harmonicBasis[peakPosition-1, basisVectorNr] == 0))
          harmonicBasis[peakPosition-1, basisVectorNr] <- harmonic^(-decay) / peakslope
        if(harmonicBasis[peakPosition+1, basisVectorNr] == 0)
          harmonicBasis[peakPosition+1, basisVectorNr] <- harmonic^(-decay) / peakslope
      }
    }
  }
  return(harmonicBasis)
}

MIR.Window.Hamming <- function(N)
{
  return(0.53836 - 0.46164 * cos(2 * pi * (0:(N-1)) / N))
}

MIR.CQT.GenerateKernel <- function(minFrequency=MIR.Convert.FrequencyFromPianokey(-48), maxFrequency=MIR.Convert.FrequencyFromPianokey(40), binsPerOctave=48, samplingFrequency=11025, threshold = 0.0054, windowingFunction=MIR.Window.Hamming, silent=FALSE)
{
  Q <- 1 / (2^(1/binsPerOctave)-1)
  K <- ceiling(binsPerOctave * log2(maxFrequency /minFrequency))
  fftLength <- 2^ceiling(log2(ceiling(Q * samplingFrequency / minFrequency)))
  if(!silent)
  {
    cat("\tQ = ", Q, "\n\tK = ", K, "\n\tFFT Length = ", fftLength, "\n\tCalculating [....................]\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b")
    percentDone <- 0
  }

  tempKernel <- rep(0, fftLength)
  CQTKernel <- matrix(0, nrow=fftLength, ncol=K)
  for(k in K:1)
  {
    len <- ceiling(Q * samplingFrequency / (minFrequency * 2^((k-1)/binsPerOctave)))
    tempKernel <- tempKernel * 0
    tempKernel[((fftLength-len)/2):((fftLength+len)/2)] <- (windowingFunction(len) / len) * exp(2i*pi*Q*(0:(len-1)) / len)
    specKernel <- fft(tempKernel)
    specKernel <- ifelse(abs(specKernel) <= threshold, 0, specKernel)
    specKernel <- Conj(specKernel) / fftLength
    CQTKernel[,k] <- specKernel
    if(!silent)
    {
      if(floor(20*(K-k)/K) > percentDone)
      {
        percentDone <- floor(20*(K-k)/K)
        cat("#")
      }
    }
  }
  if(!silent)
  {
    cat("#]\n")
  }
  return(CQTKernel)
}

MIR.CQT.CQgram <- function(x, CQTKernel, framePos, samplingFrequency=11025, hopSize=1024, windowingFunction=MIR.Window.Hamming(nrow(CQTKernel)), silent=FALSE)
{
  numberOfFrames = length(framePos)
  frameSize = nrow(CQTKernel)
  CQgram = matrix(0, ncol=numberOfFrames, nrow=dim(CQTKernel)[2])
  if(!silent)
  {
    percentDone = 0
    cat("\tFrames count = ", numberOfFrames, "\n\tCalculating [....................]\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b")
  }
  for(frameNr in 1:numberOfFrames)
  {
    dataBegin = framePos[frameNr] - frameSize/2
    dataEnd   = framePos[frameNr] + frameSize/2 - 1
    if(dataEnd > length(x))
    {
      CQGram = CQGram[,1:(frameNr-1)]
      break
    }
    CQgram[,frameNr] = abs(t(as.matrix(fft(x[dataBegin:dataEnd] * windowingFunction))) %*% CQTKernel)

    if(!silent)
    {
      if(floor(20*frameNr/numberOfFrames) > percentDone)
      {
        percentDone = floor(20*frameNr/numberOfFrames)
        cat("#")
      }
    }
  }
  cat("#]                  \n")
  return(CQgram)
}

HNMF <- function(X, A, iterations, r=1, epsilon=0.000001, silent=F)
{
  S <- matrix(abs(rnorm(ncol(X)*ncol(A))), ncol=ncol(X))
  objHistory <- c()

  for(iteration in 1:iterations)
  {
    if(!silent)
      cat("Iteration ", iteration, " ")
    X_ = A %*% S + 1e-9
    objective <- sum((X - X_)^2)
    objHistory <- c(objHistory, objective)
    if(!silent)
      cat("\t objective = ", objective, "\n")
    if(((abs(objHistory[length(objHistory)-1] - objective) / objHistory[length(objHistory)-1]) < epsilon) && (iteration > 1))
      break
    S  <- S * (t(A) %*% (X / (X_^r))) / (t(A) %*% (X_^(1-r)) + 1e-9)
    S[which(S < 0)] = 0
   }
  return(S / max(S))
}

getNMFdata = function(filename, framePos, timeBasis, frameSize, maxiter=200, peakslope=3, decay=1)
{
  if(timeBasis != 2)
    stop("Only beat-based time units are supported!")

  soundData = loadSample(paste("Wav16/", filename, ".wav", sep=""))

  paddedData = rep(0, sampleLength(soundData) + nrow(CQTkernel))
  paddedData[(nrow(CQTkernel)/2):(nrow(CQTkernel)/2 + sampleLength(soundData) - 1)] = sound(soundData)
  framePos = framePos * 16000 + nrow(CQTkernel) / 2 # convert to sample numbers

  cat(" → Calculating CQT...\n")
  X = MIR.CQT.CQgram(paddedData, CQTkernel, framePos=framePos, samplingFrequency=rate(soundData), hopSize=hopSize)

  cat(" → Calculating HNMF...\n")
  return(HNMF(X, createNMFbasis(ncol(CQTkernel), peakslope=peakslope, decay=decay), iterations=maxiter))
}

##################################################
## Main code

cat(" → Precalculating CQT kernel matrix...\n")
CQTkernel = MIR.CQT.GenerateKernel(minFrequency=MIR.Convert.FrequencyFromPianokey(1), maxFrequency=MIR.Convert.FrequencyFromPianokey(99), samplingFrequency=16000, binsPerOctave=36, threshold=0)

rwcFileDurations = read.table("Durations/durations.csv")

for(rwcfilename in allRwcFiles)
{
  cat(rep("-", 160), "\n| File name = ", rwcfilename, "\n", rep("-", 160), "\n", sep="")
  beats = read.table(paste("Beats/", rwcfilename, ".csv", sep=""), sep=",")[,1]
  thisFileDuration = rwcFileDurations[which(rwcFileDurations[,1] == paste(rwcfilename, ".wav", sep=""))[1], 2]

  framePos = c()
  for(ii in 1:(length(beats)-1))
    framePos = c(framePos, seq(from=beats[ii], to=beats[ii+1], length.out=framesPerBeat+1)[1:framesPerBeat])
  framePos = c(framePos, seq(from=beats[length(beats)], to=thisFileDuration, length.out=framesPerBeat+1)[1:framesPerBeat])

  audiodata = getNMFdata(rwcfilename, framePos=framePos, timeBasis=timeBasis, frameSize=frameSize)
  save(audiodata, file=paste("NMF/", rwcfilename, ".Rdata", sep=""))
}
