# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library("sound")

source("fileList.R")

timeBasis = 2   # don't change
frameSize = 1/6
framesPerBeat = 1/frameSize

rwcFileDurations = read.table("Durations/durations.csv")

for(analyzedFile in harmonyRwcFiles)
{
  cat(analyzedFile, ":\n", sep="")
  load(paste("NMF/", analyzedFile, ".Rdata", sep=""))
  load(paste("Processed/", analyzedFile, ".txt.Rdata", sep=""))

  # The audio can be shifted by an unknown number of frames,
  # so an average over pitches of the cross-correlation functions
  # between NMF and MIDI is calculated to determine this shift.
  ccfs = rep(NA, 88)
  for(rowNr in 1:88)
  {
    if(sum(dataMatrix[rowNr,]) > 0)
    {
      crossCorrelationFunction = ccf(audiodata[rowNr,], dataMatrix[rowNr,], plot=FALSE)
      if(length(which.max(crossCorrelationFunction$acf)) > 0)
        ccfs[rowNr] = crossCorrelationFunction$lag[which.max(crossCorrelationFunction$acf)]
    }
  }
  cat("Mean shift    =", mean(ccfs, na.rm=T), "\n")
  cat("Std deviation =", sd(ccfs, na.rm=T), "\n")

  # Adjust audio data according to the mean cross-correlation
  MIDIshift = round(mean(ccfs, na.rm=T))
  if(MIDIshift > 0)
    audiodata = audiodata[,MIDIshift:ncol(audiodata)]   # cut columns at the beginning
  else if(MIDIshift < 0)
    audiodata = cbind(matrix(0, ncol=-MIDIshift, nrow=88), audiodata)  # add empty columns

  # Cut both matrices to the shortest one
  if(ncol(audiodata) > ncol(dataMatrix))
    audiodata = audiodata[,1:ncol(dataMatrix)]
  else if(ncol(audiodata) < ncol(dataMatrix))
    dataMatrix = dataMatrix[,1:ncol(audiodata)]

  save(audiodata, file=paste("NMF/", analyzedFile, ".Rdata", sep=""))
  save(dataMatrix, file=paste("Processed/", analyzedFile, ".txt.Rdata", sep=""))
#   png(paste("Plots/Data/data_", analyzedFile, "_audio.png", sep=""), width=1600, height=1200)
#   image(t(audiodata), col=gray.colors(256,1,0,1))
#   dev.off()
#   png(paste("Plots/Data/data_", analyzedFile, "_symbolic.png", sep=""), width=1600, height=1200)
#   image(dataMatrix[,2:89], col=gray.colors(256,1,0,1))
#   dev.off()
}
