# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library('multicore')

source("Auxiliary/PRF.R")
# source("Datasets/datasets.R")

mutopiaFiles.testing = unlist(read.table("Datasets/Mutopia/mutopiaFiles.testing.txt", stringsAsFactors=FALSE), use.names=FALSE)

thresholdingResults.optimalThreshold = function(filename)
{
  cat(filename, ": loading...\n", sep="")
  processedFilename = gsub("/", "_", filename)
  processedFilename = gsub("\\.mid", "", processedFilename)
  load(paste("Datasets/Mutopia/Processed/", processedFilename, ".txt.Rdata", sep=""))
  processedFilename = gsub("\\.", "_", processedFilename)
  load(paste("Datasets/Mutopia/NMF/", processedFilename, ".Rdata", sep=""))

  if(ncol(audiodata) > ncol(dataMatrix))
  {
    cat("Size mismatch:", ncol(audiodata), "vs", ncol(dataMatrix), "\n")
    audiodata = audiodata[,1:ncol(dataMatrix)]
  }
  if(ncol(dataMatrix) > ncol(audiodata))
  {
    cat("Size mismatch:", ncol(audiodata), "vs", ncol(dataMatrix), "\n")
    dataMatrix = dataMatrix[,1:ncol(audiodata)]
  }

  cat(filename, ": thresholding...\n", sep="")
  maxS = max(audiodata)
  ii = 1
  resultingF = rep(NA, 100)
  thresholdValues = seq(0, maxS, length.out=length(resultingF))
  for(thresholdValue in thresholdValues)
  {
    resultingMatrix.threshold = ifelse(audiodata > thresholdValue, 1, 0)
    resultingF[ii] = PRF.onsets(resultingMatrix.threshold, dataMatrix)[3]
    ii = ii + 1
  }
  bestThreshold = thresholdValues[which.max(resultingF)]
  resultingMatrix.threshold = ifelse(audiodata > bestThreshold, 1, 0)
  evaluationResults = c(ncol(dataMatrix), bestThreshold, PRF.onsets(resultingMatrix.threshold, dataMatrix))

  save(evaluationResults, file=paste("Results/thresholding_", processedFilename, ".Rdata", sep=""))
  cat(filename, ": PRF =", evaluationResults, "\n")
  return(evaluationResults)
}

thresholdingResults.givenThreshold = function(filename, threshold)
{
  processedFilename = gsub("/", "_", filename)
  processedFilename = gsub("\\.mid", "", processedFilename)
  load(paste("Datasets/Mutopia/Processed/", processedFilename, ".txt.Rdata", sep=""))
  processedFilename = gsub("\\.", "_", processedFilename)
  load(paste("Datasets/Mutopia/NMF/", processedFilename, ".Rdata", sep=""))

  if(ncol(audiodata) > ncol(dataMatrix))
    audiodata = audiodata[,1:ncol(dataMatrix)]
  if(ncol(dataMatrix) > ncol(audiodata))
    dataMatrix = dataMatrix[,1:ncol(audiodata)]

  resultingMatrix.threshold = ifelse(audiodata > threshold, 1, 0)
  evaluationResults = c(PRF.onsets(resultingMatrix.threshold, dataMatrix))

  cat(evaluationResults[3], " ", sep="")
  return(evaluationResults)
}

thresholdingResults.varyingThreshold = function(filename, thresholds)
{
  processedFilename = gsub("/", "_", filename)
  processedFilename = gsub("\\.mid", "", processedFilename)
  load(paste("Datasets/Mutopia/Processed/", processedFilename, ".txt.Rdata", sep=""))
  processedFilename = gsub("\\.", "_", processedFilename)
  load(paste("Datasets/Mutopia/NMF/", processedFilename, ".Rdata", sep=""))

  if(ncol(audiodata) > ncol(dataMatrix))
    audiodata = audiodata[,1:ncol(dataMatrix)]
  if(ncol(dataMatrix) > ncol(audiodata))
    dataMatrix = dataMatrix[,1:ncol(audiodata)]

  resultingMatrix.threshold = matrix(NA, nrow=nrow(audiodata), ncol=ncol(audiodata))
  for(pitchNr in 1:88)
    resultingMatrix.threshold[pitchNr,] = ifelse(audiodata[pitchNr,] > thresholds[pitchNr], 1, 0)

  evaluationResults = c(PRF.onsets(resultingMatrix.threshold, dataMatrix))

  cat(evaluationResults[3], " ", sep="")
  return(evaluationResults)
}

getSingleResults = function(dataset, threshold)
{
  cat(paste(rep("-", 120), collapse=""), "\n")
  cat("Threshold =", threshold, "\n\t")
  fmeasure = mean(matrix(unlist(mclapply(dataset, function(filename) thresholdingResults.givenThreshold(filename, threshold=threshold))), nrow=3)[3,])
  cat("\nF-measure =", fmeasure, "\n")
  return(fmeasure)
}
