# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

# library("multicore")

source("Datasets/datasets.R")
source("Auxiliary/PRF.R")
source("Settings/training.R")
source("Settings/audio.R")
source("Training/probabilityDefinitions.R")
chordPty = log(chordPty)

PRFfunction = PRF.onsets
#PRFfunction = PRF

#obtainResults = function(acousticFudgeFactors, bernoulliFudgeFactor, suffix, fileNrs)
#obtainResults = function(acousticFudgeFactor, bernoulliFudgeFactors, suffix, fileNrs)
#obtainResults = function(acousticFudgeFactor, bernoulliFudgeFactor, suffix, fileNr)
obtainResults = function(acousticFudgeFactors, bernoulliFudgeFactors, suffix, fileNr)
{
  filename = paste("Temp/probArrayEnv", fileNr, ".", suffix, ".Renv", sep="")
  cat(fileNr, " (", filename, "): ", "Loading precalculated data...\n", sep="")
  load(filename)
  quantizedS = ceiling(S * quantizationLevels)
  allResults = array(NA, dim=c(length(acousticFudgeFactors), length(bernoulliFudgeFactors), 3))
  for(ii in seq_along(bernoulliFudgeFactors))
  {
    bernoulliFudgeFactor = bernoulliFudgeFactors[ii]
    allResults[,ii,] = sapply(acousticFudgeFactors, function(acousticFudgeFactor) {
      load(file="TrainedParameters/onProbabilities.Rdata")
      load(file="TrainedParameters/offProbabilities.Rdata")
      onProbabilities  = acousticFudgeFactor * log(onProbabilities)
      offProbabilities = acousticFudgeFactor * log(offProbabilities)
      source("Experiments.Audio/precalculateOutputMatrix.R", local=TRUE)
      source("Experiments.Audio/performModifiedFrontierAlgorithm.R", local=TRUE)
      save(resultingMatrix, file=paste("Results/resultingMatrix_", fileNr, "_", acousticFudgeFactor, "_", bernoulliFudgeFactor, ".Rdata", sep=""))
      PRFresults = PRFfunction(resultingMatrix, dataMatrix[,1:ncol(S)])[3]
      cat("(", fileNr, "/", acousticFudgeFactor, "/", bernoulliFudgeFactor, ") ", " = ", PRFresults, "\n", sep="")
      return(PRFresults)
    })
    cat(allResults[,ii,], "\n")
  }
  return(allResults)
}

plotRF = function(results, groundtruth, filename)
{
  plotMatrix = results * 0
  plotMatrix[(results == 0) & (groundtruth == 0)] = 0  # true negative
  plotMatrix[(results > 0) & (groundtruth > 0)] = 1    # true positive
  plotMatrix[(results > 0) & (groundtruth == 0)] = 2   # false positive
  plotMatrix[(results == 0) & (groundtruth > 0)] = 3   # false negative
  pdf(filename, width=10, height=5)
  image(x=1:ncol(plotMatrix), y=1:88, z=t(plotMatrix), col=c("white", "black", "red", "blue"), xlab="Time frame", ylab="Piano key nr")
  dev.off()
}

#parameterValues = seq(0.95,1,length.out=51)
#for(ii in 1:51)
#{
#  cat("Running for ii =", ii, "\n")
#  singleResult = obtainResults(1,parameterValues[ii],101)
#  save(singleResult, file=paste("Results/fixedAcoustic_", ii, ".Rdata", sep=""))
#}

#singleResult = obtainResults(1,seq(0,1,length.out=51),1001,c(1:6,8:10,12))
#singleResult = obtainResults(1,seq(0,100,length.out=21),1001,c(1:20))
#singleResult = obtainResults(1,0.9,1001,1)
#singleResult = obtainResults(seq(0.1,2,length.out=21),1,1001,c(1,3,10,17))

#pdf("results_onlyD.pdf", width=8, height=6)
#plot(x=seq(0.1,2,length.out=21), y=rowMeans(singleResult), type="l", ylim=c(0,1))
#dev.off()

