# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

##################################################################################################
## Determining optimal state sequence (with Frontier algorithm)

D  = array(0, dim=c(24, extNumberOfCombinations))
DI = list()
A  = array(0, dim=c(24, extNumberOfCombinations, 24))
B  = array(0, dim=c(extNumberOfCombinations, 24))
BI = list()
C  = array(0, dim=c(extNumberOfCombinations, 24, extNumberOfCombinations))

ProbNotesGivenChord = function(noteActivity, chordNr)
{
  notesBelow = (c(rev(cumsum(rev(noteActivity))), 0) + 1)[2:89]  # mysterious R magic...
  if(all(is.na(polyphonyPties.fudged)))
    polyphonyPties.fudged[] = 1
  if(all(is.na(harmonyPties.fudged)))
    harmonyPties.fudged[] = 1
  if(all(is.na(voicePties.fudged)))
    voicePties.fudged[] = 1
  unnormalizedProbs = t(harmonyPties.fudged[, chordNr, ]) * polyphonyPties.fudged[notesBelow, ] * voicePties.fudged[rep(89, 88), ]
  normalizedProbs = unnormalizedProbs / rowSums(unnormalizedProbs) # normalization
  return(sum(log(normalizedProbs[cbind(1:88, noteActivity + 1)])))
}

# First time slice
currIndices <- getCombinationIndices(1, indexTranslator)
for(n0 in 1:extNumberOfCombinations)
{
  for(c0 in 1:24)
    D[c0, n0] = log2(1/24) + (1-bernoulliFudgeFactor)*ProbNotesGivenChord(reducedCombinationMatrix[currIndices[n0],], c0) + outputPtiesFinal[n0, 1]
#    D[c0, n0] = log2(1/24) + probArray[cbind(currIndices, 1:24, lastIndices)]
}

# Middle time slices
for(timeSlice in 2:ncol(S))
{
  # We always start with the frontier consisting of the previous note combination
  # and the previous chord: D = P(ct_, nt_)
  
  lastIndices = currIndices
  currIndices = getCombinationIndices(timeSlice, indexTranslator)
  
  # Include the current chord in the frontier: A = D·P(ct | ct_) = P(ct_, nt_, ct)
  for(nt_ in 1:extNumberOfCombinations)
    for(ct in 1:24)
      A[, nt_, ct] = D[,nt_] + chordPty[ct,]
  
  # Integrate the previous chord out of the frontier: B = Σ_{ct_} A = P(nt_, ct)
  BI[[timeSlice-1]] = array(0, dim=c(extNumberOfCombinations, 24)) # save the indices of the maxima
  for(nt_ in 1:extNumberOfCombinations)
    for(ct in 1:24)
    {
      whichMax = which.max(A[, nt_, ct])
      BI[[timeSlice-1]][nt_, ct] = whichMax
      B[nt_, ct] = A[whichMax, nt_, ct]
    }
  
  # Include the current note salience (and immediately remove it, and then
  # the current note combination in the frontier: C = B·P(nt | ct, nt_) = P(nt_, ct, nt)
  for(nt_ in 1:extNumberOfCombinations)
    for(ct in 1:24)
      C[nt_,ct,] = B[nt_,ct] + outputPtiesFinal[, timeSlice] + probArray[currIndices,ct,lastIndices[nt_]]
  
  # Integrate the previous note combination out of the frontier: D = Σ_{nt_} C = P(ct, nt)
  DI[[timeSlice-1]] = array(0, dim=c(24, extNumberOfCombinations))
  for(ct in 1:24)
    for(nt in 1:extNumberOfCombinations)
    {
      whichMax = which.max(C[, ct, nt])
      DI[[timeSlice-1]][ct, nt] = whichMax
      D[ct, nt] = C[whichMax, ct, nt]
    }
}

# Last time slices
EI = rep(0, extNumberOfCombinations)
E  = rep(0, extNumberOfCombinations)
for(nT in 1:extNumberOfCombinations)
{
  EI[nT] = which.max(D[, nT])
  E[nT]  = D[EI[nT], nT]
}
FI = which.max(E)
F  = E[FI]

# Backtracing
hiddenNodes <- matrix(NA, 2, ncol(S))
combinationNrs <- rep(NA, ncol(S))
hiddenNodes[2, ncol(S)] <- FI
hiddenNodes[1, ncol(S)] <- EI[FI]
combinationIndices <- getCombinationIndices(ncol(S), indexTranslator)
combinationNrs[ncol(S)] <- combinationIndices[FI]
for(timeSlice in (ncol(S) - 1):1)
{
	hiddenNodes[2, timeSlice] <- DI[[timeSlice]][hiddenNodes[1, timeSlice + 1], hiddenNodes[2, timeSlice + 1]]
	hiddenNodes[1, timeSlice] <- BI[[timeSlice]][hiddenNodes[2, timeSlice], hiddenNodes[1, timeSlice + 1]]
	combinationIndices <- getCombinationIndices(timeSlice, indexTranslator)
	combinationNrs[timeSlice] <- combinationIndices[hiddenNodes[2, timeSlice]]
}
resultingMatrix = t(reducedCombinationMatrix[combinationNrs,])
