# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library("gtools")
library("multicore")

##################################################################################################
## Creating list of all candidate note sequences:

source("Auxiliary/rightSmoothing.R", local=TRUE)
quantizedS = ceiling(S * quantizationLevels)

if(includeRealSequence)
  numberOfCombinations = numberOfCombinations + 1
extNumberOfCombinations = (includePreviousFrames + 1) * numberOfCombinations

cat(filename, ": finding the biggest peaks...\n", sep="")
peakMatrix <- matrix(0, ncol(S), numberOfPeakCandidates)
for(timeSlice in 1:ncol(S))
{
  peakMatrix[timeSlice, ] = sort(smoothedS[, timeSlice], index.return=T, decreasing=T)$ix[1:numberOfPeakCandidates]
  peakMatrix[timeSlice, ] = ifelse(smoothedS[peakMatrix[timeSlice, ], timeSlice] > minimalSelectedSValue, peakMatrix[timeSlice, ], NA)
}

# image(x=1:ncol(S), y=1:88, z=t(smoothedS), col=gray.colors(256,1,0,1))
# for(timeSlice in 1:ncol(S))
#   points(x=rep(timeSlice, numberOfPeakCandidates), y=peakMatrix[timeSlice, ], col="red", pch=17)
# X11()
# image(x=1:ncol(S), y=1:88, z=t(dataMatrix[,1:ncol(S)]), col=gray.colors(256,1,0,1))
# for(timeSlice in 1:ncol(S))
#   points(x=rep(timeSlice, numberOfPeakCandidates), y=peakMatrix[timeSlice, ], col="red", pch=17)
# 
# realNoteRatio = sum(sapply(1:ncol(S), function(timeSlice) sum(dataMatrix[peakMatrix[timeSlice, ],timeSlice] == 1))) * 100 / sum(dataMatrix[,1:ncol(S)])

cat(filename, ": creating a list of all note combination candidates...\n", sep=""); #cat(system.time({
combinationMatrix = matrix(0, ncol(S) * numberOfCombinations + 1, 88)    # This will hold a list of possible combinations used throughout analysis
for(timeSlice in 1:ncol(S))
{
  combNr = 1
  allCombinationMatrix = matrix(1, 2^numberOfPeakCandidates - 1, 88)  # we fill it with 1s so that inexisting combinations get extremely small probabilities later
  sequenceFitness = rep(0, nrow(allCombinationMatrix))
  currentSelectedNotes = peakMatrix[timeSlice, !is.na(peakMatrix[timeSlice, ])]
  if(length(currentSelectedNotes) > 0)
    for(combSize in 1:length(currentSelectedNotes))
    {
      noteIndices = combinations(length(currentSelectedNotes), combSize, currentSelectedNotes)
      for(ii in 1:nrow(noteIndices))
      {
        allCombinationMatrix[combNr, -noteIndices[ii,]] = 0
        combNr = combNr + 1
      }
    }
  for(ii in 1:nrow(allCombinationMatrix))
  {
    currNotes = which(allCombinationMatrix[ii, ] > 0)
  #       sequenceFitness[combNr] = sum(smoothedS[currNotes, timeSlice]) / (sum(smoothedS[, timeSlice]) * (combSize^0.65))
    if(length(currNotes) == 0)
      sequenceFitness[ii] = sum(offProbability[quantizedS[, timeSlice]])
    else
      sequenceFitness[ii] = sum(onProbability[quantizedS[currNotes, timeSlice]], offProbability[quantizedS[-currNotes, timeSlice]])
  }
  if(includeRealSequence)
  {
    if(numberOfCombinations > 1)
      combinationMatrix[((timeSlice-1)*numberOfCombinations + 1):(timeSlice * numberOfCombinations), ] = rbind(allCombinationMatrix[sort(sequenceFitness, index.return=T, decreasing=T)$ix[1:(numberOfCombinations-1)], ], dataMatrix[, timeSlice])
    else
      combinationMatrix[((timeSlice-1)*numberOfCombinations + 1):(timeSlice * numberOfCombinations), ] = dataMatrix[, timeSlice]
  }
  else
    combinationMatrix[((timeSlice-1)*numberOfCombinations + 1):(timeSlice * numberOfCombinations), ] = allCombinationMatrix[sort(sequenceFitness, index.return=T, decreasing=T)$ix[1:numberOfCombinations], ]
}
cat(filename, ": found ", nrow(combinationMatrix), " combinations\n", sep="")

##################################################################################################
## Removing repeated combinations
cat(filename, ": finding unique note combinations...\n", sep="")
reducedCombinationMatrix <- unique(combinationMatrix)
indexTranslator <- rep(nrow(reducedCombinationMatrix), nrow(combinationMatrix))  # mapping of all indices to reduced ones
combinationLengths <- rowSums(reducedCombinationMatrix)
maxArraySize = (nrow(reducedCombinationMatrix)^2)*24*8/(1024^3)  # 24 chords, one 'double' value is 8 bytes
cat(filename, ": found ", nrow(reducedCombinationMatrix), " unique combinations, note transition array will have ", round(maxArraySize, 3), "GB\n", sep="")

if(maxArraySize > memoryLimit)
  stop("Probability array above the set memory limit!")

cat(filename, ": mapping full combination matrix to the reduced one...\n", sep=""); #cat(system.time({
  fullCombinationLengths <- rowSums(combinationMatrix)
  combinationMatrix <- t(combinationMatrix)
  for(ii in 1:nrow(reducedCombinationMatrix))
  {
    sameLengths <- which((fullCombinationLengths == combinationLengths[ii]))
    indexTranslator[sameLengths[which(colSums(as.matrix(combinationMatrix[,sameLengths] == reducedCombinationMatrix[ii,])) == 88)]] <- ii
  }
  combinationMatrix <- t(combinationMatrix)
  reducedCombinationMatrix <- reducedCombinationMatrix > 0
# })[3], "seconds\n")

##################################################################################################
## Probability matrix precalculation

# This function returns the indices of note combinations chosen for the given time frame
getCombinationIndices = function(timeSlice, indexTranslator)
{
  fromNr = (timeSlice - 1) * numberOfCombinations + 1 - includePreviousFrames*numberOfCombinations
  if(fromNr < 1)
  {
    negativeIndexCount = 1 - fromNr
    fromNr = 1
  }
  else
    negativeIndexCount = 0
  toNr   = timeSlice * numberOfCombinations
  
  return(indexTranslator[c(rep(fromNr, negativeIndexCount),  seq(from=fromNr, to=toNr))])
}

uninformedPty = 2^-88
solutionSpaceSize = nrow(reducedCombinationMatrix)

#load(paste("TrainedParameters/fudgeFactors.HCDVPN.loglinear.Rdata", sep=""))
#fudgeFactorValues = fudgeFactors$par
#load(paste("TrainedParameters/fudgeFactors.HCV.loglinear.Rdata", sep=""))
#fudgeFactorValues = c(fudgeFactors$par[1], 0, fudgeFactors$par[2], 0, 0)
#load(paste("TrainedParameters/fudgeFactors.HC.loglinear.Rdata", sep=""))
#fudgeFactorValues = c(fudgeFactors$par[1], 0, 0, 0, 0)
load(paste("TrainedParameters/fudgeFactors.D.loglinear.Rdata", sep=""))
fudgeFactorValues = c(0, fudgeFactors$par[1], 0, 0, 0)
#load(paste("TrainedParameters/fudgeFactors.V.loglinear.Rdata", sep=""))
#fudgeFactorValues = c(0, 0, fudgeFactors$par[1], 0, 0)
#load(paste("TrainedParameters/fudgeFactors.N.loglinear.Rdata", sep=""))
#fudgeFactorValues = c(0, 0, 0, 0, fudgeFactors$par[1])
#load(paste("TrainedParameters/fudgeFactors.P.loglinear.Rdata", sep=""))
#fudgeFactorValues = c(0, 0, 0, fudgeFactors$par[1], 0)

harmonyPties.fudged = harmonyPties^fudgeFactorValues[1]
durationPties.fudged = durationPties^fudgeFactorValues[2]
voicePties.fudged = simpleVoiceProbs^fudgeFactorValues[3]
polyphonyPties.fudged = polyphonyPties^fudgeFactorValues[4]
neighbourPties.fudged = neighbourPties^fudgeFactorValues[5]

cat(filename, ": calculating numbers of notes below...\n", sep="")
notesBelow = matrix(unlist(lapply(1:solutionSpaceSize, function(frameNr) {
  return((c(rev(cumsum(rev(reducedCombinationMatrix[frameNr, ]))), 0) + 1)[2:89])  # mysterious R magic...
})), nrow=88)

# Only small number of transitions (~15%) needs to be calculated
cat(filename, ": chosing note combinations to calculate... ", sep="")
possibleNoteTransitions = matrix(FALSE, solutionSpaceSize, solutionSpaceSize)
indexMatrix = matrix(NA, numberOfCombinations, solutionSpaceSize)
for(frameNr in 1:solutionSpaceSize)
  indexMatrix[,frameNr] = getCombinationIndices(frameNr, indexTranslator)
for(frameNrCurr in 1:solutionSpaceSize)
{
  if((frameNrCurr %% 100) == 0)
    cat(frameNrCurr, " ", sep="")
  for(frameNrPrev in 1:solutionSpaceSize)
    possibleNoteTransitions[cbind(indexMatrix[,frameNrPrev], indexMatrix[,frameNrCurr])] = TRUE
}
cat(" we'll need to calculate", round(100*length(which(possibleNoteTransitions))/prod(dim(possibleNoteTransitions)),2), "% of combinations\n", sep="")

cat(filename, ": calculating note combination transition probabilities...\n");
probArrays = mclapply(1:24, function(chordNr) {   # use mclaplly for multicore
  distances = rep(NA, 88)
  probArray = array(-Inf, dim=c(solutionSpaceSize, solutionSpaceSize))
  unnormalizedProbs.H = t(harmonyPties.fudged[, chordNr, ])
  for(frameNrCurr in 1:solutionSpaceSize)
  {
    if((frameNrCurr %% 100) == 0)
      cat(filename, ": frame ", frameNrCurr, " of ", solutionSpaceSize, "\n", sep="")
    unnormalizedProbs.HPN = unnormalizedProbs.H * polyphonyPties.fudged[notesBelow[,frameNrCurr], ] * t(neighbourPties[,c(1, reducedCombinationMatrix[frameNrCurr, 1:87] + 1)])
    for(frameNrPrev in which(possibleNoteTransitions[,frameNrCurr]))
    {
      notesPrev = which(reducedCombinationMatrix[frameNrPrev, ])
      if(length(notesPrev) > 0)
        distances = apply(abs(outer(1:88, notesPrev, "-")), FUN=min, MARGIN=1) + 1
      else
        distances = rep(89, 88)

      unnormalizedProbs.HPNDV = unnormalizedProbs.HPN * voicePties.fudged[distances, ] * cbind(durationPties.fudged[cbind(1, 1:88, reducedCombinationMatrix[frameNrPrev,] + 1)], durationPties.fudged[cbind(2, 1:88, reducedCombinationMatrix[frameNrPrev,] + 1)])
#        unnormalizedProbs = cbind(durationPties.fudged[cbind(1, 1:88, reducedCombinationMatrix[frameNrPrev,] + 1)], durationPties.fudged[cbind(2, 1:88, reducedCombinationMatrix[frameNrPrev,] + 1)])
#        unnormalizedProbs = cbind(priorOffProbabilities, priorOnProbabilities)
      normalizedProbs = log(unnormalizedProbs.HPNDV / rowSums(unnormalizedProbs.HPNDV)) # normalization
      probArray[frameNrCurr, frameNrPrev] = sum(normalizedProbs[cbind(1:88, reducedCombinationMatrix[frameNrCurr, ] + 1)])
    }
  }
  return(probArray)
})
# Collect results from [mc]lapply:
probArray = array(NA, dim=c(solutionSpaceSize, 24, solutionSpaceSize)) # copy the results to the global probability array
for(chordNr in 1:24)
  probArray[,chordNr,] = probArrays[[chordNr]]

