# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library("sound")

timeBasis = 2
frameSize = 1/6

##WARNING: I assume that the symbolic data have been prepared with timeBasis=2 and identical frameSize as used for audio data
auftaktLength = function(filename)
{
  load(paste("Data/Processed/", filename, ".Rdata", sep=""))
  return(which(cumsum(rowSums(dataMatrix[,2:89])) > 0)[1] - 1)
}

samplesPerFrame = function(filename, frameSize, sampleRate)
{
  bpms = read.table("Data/BPM.txt")
  if(!any(bpms[,1] == filename))
    stop("No information about tempo for file ", filename)
  bpm = bpms[which(bpms[,1] == filename)[1], 2]  # beats per minute
  return(sampleRate * (60.0 / bpm) * frameSize) # time unit length in seconds
}

MIR.Convert.FrequencyFromMIDInote <- function(note) return(440 * 2 ^ (note / 12))
MIR.Convert.FrequencyFromPianokey <- function(pianokey) MIR.Convert.FrequencyFromMIDInote(pianokey - 49)

createNMFbasis = function(kernelSize, peakslope=3, decay=1)
{
  PianokeyFromFrequency <- function(x) 12 * log(x/440., 2) + 49
  harmonicBasis <- matrix(0, kernelSize, 88)
  for(basisVectorNr in 1:ncol(harmonicBasis))
  {
    baseFrequency <- MIR.Convert.FrequencyFromPianokey(basisVectorNr)
    for(harmonic in 1:100)
    {
      peakPosition  <- (PianokeyFromFrequency(baseFrequency * harmonic) - 1) * 3 + 1
      if(peakPosition < nrow(harmonicBasis) - 1)
      {
        harmonicBasis[peakPosition, basisVectorNr] <- harmonic^(-decay)
        if((peakPosition > 1) && (harmonicBasis[peakPosition-1, basisVectorNr] == 0))
          harmonicBasis[peakPosition-1, basisVectorNr] <- harmonic^(-decay) / peakslope
        if(harmonicBasis[peakPosition+1, basisVectorNr] == 0)
          harmonicBasis[peakPosition+1, basisVectorNr] <- harmonic^(-decay) / peakslope
      }
    }
  }
  return(harmonicBasis)
}

MIR.Window.Hamming <- function(N)
{
  return(0.53836 - 0.46164 * cos(2 * pi * (0:(N-1)) / N))
}

MIR.CQT.GenerateKernel <- function(minFrequency=MIR.Convert.FrequencyFromPianokey(-48), maxFrequency=MIR.Convert.FrequencyFromPianokey(40), binsPerOctave=48, samplingFrequency=11025, threshold = 0.0054, windowingFunction=MIR.Window.Hamming, silent=FALSE)
{
  Q <- 1 / (2^(1/binsPerOctave)-1)
  K <- ceiling(binsPerOctave * log2(maxFrequency /minFrequency))
  fftLength <- 2^ceiling(log2(ceiling(Q * samplingFrequency / minFrequency)))
  if(!silent)
  {
    cat("\tQ = ", Q, "\n\tK = ", K, "\n\tFFT Length = ", fftLength, "\n\tCalculating [....................]\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b")
    percentDone <- 0
  }

  tempKernel <- rep(0, fftLength)
  CQTKernel <- matrix(0, nrow=fftLength, ncol=K)
  for(k in K:1)
  {
    len <- ceiling(Q * samplingFrequency / (minFrequency * 2^((k-1)/binsPerOctave)))
    tempKernel <- tempKernel * 0
    tempKernel[((fftLength-len)/2):((fftLength+len)/2)] <- (windowingFunction(len) / len) * exp(2i*pi*Q*(0:(len-1)) / len)
    specKernel <- fft(tempKernel)
    specKernel <- ifelse(abs(specKernel) <= threshold, 0, specKernel)
    specKernel <- Conj(specKernel) / fftLength
    CQTKernel[,k] <- specKernel
    if(!silent)
    {
      if(floor(20*(K-k)/K) > percentDone)
      {
        percentDone <- floor(20*(K-k)/K)
        cat("#")
      }
    }
  }
  if(!silent)
  {
    cat("#]\n")
  }
  return(CQTKernel)
}

MIR.CQT.CQgram <- function(x, CQTKernel, samplingFrequency=11025, hopSize=1024, windowingFunction=MIR.Window.Hamming(nrow(CQTKernel)), silent=FALSE)
{
  numberOfFrames <- floor((length(x) - nrow(CQTKernel)) / hopSize) + 1
#   frameNr <- 1
  CQgram <- matrix(0, ncol=numberOfFrames, nrow=dim(CQTKernel)[2])
  if(!silent)
  {
    percentDone <- 0
    cat("\tFrames count = ", numberOfFrames, "\n\tCalculating [....................]\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b")
  }
  for(frameNr in 1:numberOfFrames)
  {
    dataBegin <- round((frameNr-1)*hopSize + 1)
    dataEnd   <- round((frameNr-1)*hopSize + dim(CQTKernel)[1])
    if(dataEnd > length(x))
    {
      #frameNr <- frameNr - 1
      CQGram <- CQGram[,1:(frameNr-1)]
      break
    }
    CQgram[,frameNr] <- abs(t(as.matrix(fft(x[dataBegin:dataEnd] * windowingFunction))) %*% CQTKernel)

    if(!silent)
    {
      if(floor(20*frameNr/numberOfFrames) > percentDone)
      {
        percentDone <- floor(20*frameNr/numberOfFrames)
        cat("#")
      }
    }
  }
  cat("#]                  \n")
  return(CQgram)
}

HNMF <- function(X, A, iterations, r=1, epsilon=0.000001, silent=F)
{
  S <- matrix(abs(rnorm(ncol(X)*ncol(A))), ncol=ncol(X))
  objHistory <- c()

  for(iteration in 1:iterations)
  {
    if(!silent)
      cat("Iteration ", iteration, " ")
    X_ = A %*% S + 1e-9
    objective <- sum((X - X_)^2)
    objHistory <- c(objHistory, objective)
    if(!silent)
      cat("\t objective = ", objective, "\n")
    if(((abs(objHistory[length(objHistory)-1] - objective) / objHistory[length(objHistory)-1]) < epsilon) && (iteration > 1))
      break
    S  <- S * (t(A) %*% (X / (X_^r))) / (t(A) %*% (X_^(1-r)) + 1e-9)
    S[which(S < 0)] = 0
   }
  return(S / max(S))
}

getNMFdata = function(filename, timeBasis, frameSize, maxiter=200, peakslope=3, decay=1)
{
  if(timeBasis != 2)
    stop("Only beat-based time units are supported!")

  soundData = loadSample(paste("Data/Wav16/", filename, ".wav", sep=""))

  cat(" → Precalculating CQT kernel matrix...\n")
  CQTkernel = MIR.CQT.GenerateKernel(minFrequency=MIR.Convert.FrequencyFromPianokey(1), maxFrequency=MIR.Convert.FrequencyFromPianokey(99), samplingFrequency=rate(soundData), binsPerOctave=36, threshold=0)

  paddedData = rep(0, sampleLength(soundData) + nrow(CQTkernel))
  paddedData[(nrow(CQTkernel)/2):(nrow(CQTkernel)/2 + sampleLength(soundData) - 1)] = sound(soundData)

  cat(" → Calculating CQT...\n")
  hopSize = samplesPerFrame(filename, frameSize=frameSize, sampleRate=rate(soundData))
  auftakt = auftaktLength(filename)
  cat("        Hop size = ", hopSize, " (", 1000*hopSize/rate(soundData), " ms; ", hopSize, " samples)\n", sep="")
  cat("        Tempo = ", 60/(hopSize/(rate(soundData)*frameSize)), " BPM\n", sep="")
  cat("        Auftakt = ", auftakt, " frames (", 1000*auftakt*hopSize/rate(soundData), " ms)\n", sep="")
  X = MIR.CQT.CQgram(paddedData, CQTkernel, samplingFrequency=rate(soundData), hopSize=hopSize)
  X = cbind(matrix(0, nrow(X), auftaktLength(filename)), X)   # add the initial rest (removed by Timidity)

  cat(" → Calculating HNMF...\n")
  return(HNMF(X, createNMFbasis(ncol(CQTkernel), peakslope=peakslope, decay=decay), iterations=maxiter))
}

##################################################
## Main code
for(filename in filenames)
{
  cat(rep("-", 160), "\n| File name = ", filename, "\n", rep("-", 160), "\n", sep="")
  audiodata = getNMFdata(filename, timeBasis=timeBasis, frameSize=frameSize)
  save(audiodata, file=paste("Data/NMF/", filename, ".Rdata", sep=""))
}

for(analyzedFile in filenames)
{
  cat(analyzedFile, ":\n", sep="")
  load(paste("Data/NMF/", analyzedFile, ".Rdata", sep=""))
  load(paste("Data/Processed/", analyzedFile, ".Rdata", sep=""))

  cat("  Audio:    ", ncol(audiodata), "\n", sep="")
  cat("  Symbolic: ", nrow(dataMatrix), "\n", sep="")

  if(ncol(audiodata) + 1 > nrow(dataMatrix))
    audiodata = audiodata[,2:(nrow(dataMatrix)+1)]
  else
    audiodata = cbind(audiodata[,2:ncol(audiodata)], matrix(0, nrow(audiodata), nrow(dataMatrix) - ncol(audiodata)   + 1))

  save(audiodata, file=paste("Data/NMF/", analyzedFile, ".Rdata", sep=""))
  png(paste("Plots/Data/data_", analyzedFile, "_audio.png", sep=""), width=1600, height=1200)
  image(t(audiodata), col=gray.colors(256,1,0,1))
  dev.off()
  png(paste("Plots/Data/data_", analyzedFile, "_symbolic.png", sep=""), width=1600, height=1200)
  image(dataMatrix[,2:89], col=gray.colors(256,1,0,1))
  dev.off()
}
