# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

source("Training/probabilityDefinitions.R")
# load("TrainedParameters/priorOnProbability.Rdata")
# load(file="TrainedParameters/polyphonyModel.Rdata")
# load(file="Results/globalNotePties.Rdata")
load(file="TrainedParameters/harmonyModel.major.Rdata")
load(file="TrainedParameters/harmonyModel.minor.Rdata")

blackPianoKeys = c(rep(12 * c(0,1,2,3,4,5,6), each=5) + c(2,5,7,10,12), 86)
whitePianoKeys = c(rep(12 * c(0,1,2,3,4,5,6), each=7) + c(1,3,4,6,8,9,11), 85, 87, 88)
barcolors = rep("white", 88)
barcolors[blackPianoKeys] = rgb(0.2, 0.2, 0.2)

## Prior note model
pdf("Plots/globalNotePty.pdf", width=10/2.54, height=4/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(priorOnProbabilities, col=barcolors, names.arg=paste(1:88+20), xlab="MIDI number", ylab="Probability", space=0)
dev.off()

## Harmony submodel
intervalNames = c("R", "m2", "M2", "m3", "M3", "P4", "T", "P5", "m6", "M6", "m7", "M7")
pdf("Plots/harmonyModel.simple.pdf", width=imageWidth/2.54, height=6/2.54)
par(cex=0.5, lwd=0.5)
  matplot(cbind(majorPties, minorPties), lty=1, type="b", pch=c("M", "m"), axes=F, xlab="Interval from root", ylab="Activation probability")
  axis(1, at=1:12, labels=intervalNames)
  axis(2)
  text(1:12, -0.06, paste(0:11), xpd=TRUE)
dev.off()

## Duration submodel
pdf("Plots/durationModel_onset.pdf", width=10/2.54, height=4/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(durationPties[2,,1], col=barcolors, names.arg=paste(1:88+20), xlab="MIDI number", ylab="Probability", space=0, ylim=c(0, 0.03))
#  legend(5, 0.8, c("silence to silence", "note onset", "note offset", "note to note"), fill=c("dark gray", "light gray", "dark red", "red"), bg="white")
#   points(x=blackPianoKeys-0.5, y=rep(1.021, length(blackPianoKeys)), pch=25, bg="black", col="black", xpd=TRUE)
  #   points(x=whitePianoKeys-0.5, y=rep(1.03, length(whitePianoKeys)), pch=24, col="black", xpd=TRUE)
dev.off()
pdf("Plots/durationModel_note.pdf", width=10/2.54, height=4/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(durationPties[2,,2], col=barcolors, names.arg=paste(1:88+20), xlab="MIDI number", ylab="Probability", space=0, ylim=c(0,1))
dev.off()

## Polyphony submodel
pdf("Plots/noteCombLength.pdf", width=10/2.54, height=6/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(noteCombinationLengthProbabilities[1:24], names.arg=0:23, ylab="Activation probability", xlab="Number of active notes lower than the current pitch")
  abline(h=priorOnProbability, col="dark red", lty=2, lwd=1)
dev.off()

## Neighbor submodel
pdf("Plots/neighborModel.pdf", width=10/2.54, height=6/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(neighbourPties[2,], names.arg=c("0,0,1", "0,1,1", "1,0,1", "1,1,1"), ylab="Activation probability", xlab="Succesion of pitches")
  abline(h=priorOnProbability, col="dark red", lty=2, lwd=1)
dev.off()

## Voice submodel
octaveColors = c(rep("#808080", 6), "white", rep("#808080", 4), "#c0c0c0")
pdf("Plots/voiceModel.pdf", width=imageWidth/2.54, height=6/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(c(simpleVoiceProbs[1:39,2], rep(NA,5), simpleVoiceProbs[76:89,2]), pch=20, xlab="Distance to the previous closest active note", ylab="Activation probability (log)", names.arg=c(0:38, rep("", 5), 75:87, "+Inf"), log="y", ylim=c(1e-5, 1), col=c("#c0c0c0", rep(octaveColors, 3), rep("#808080", 20), "black"))
  axis(1, at=58*1.2-0.55, label=expression(+infinity), tick=FALSE, col="black", las=0, padj=-0.5, hadj=0.35)
  text(41.5*1.2, 1e-4, "...")
  load("TrainedParameters/priorOnProbability.Rdata")
  abline(h=priorOnProbability, col="dark red", lty=2, lwd=1)
dev.off()

# Salience model
quantizationLevels <- 500
load(file="TrainedParameters/onProbability.Rdata")
load(file="TrainedParameters/offProbability.Rdata")
crossPointX = which(onProbability > offProbability)[1]
crossPointY = onProbability[crossPointX]
crossPointX = seq(0,0.04,length.out=quantizationLevels)[crossPointX]
pdf(paste("Plots/salienceProbs.pdf", sep=""), width=imageWidth/2.54, height=6/2.54)
  par(cex=0.5)
  matplot(x=seq(0,0.04,length.out=quantizationLevels), y=cbind(onProbability, offProbability), type="l", ylab="Likelihood of activity (log)", xlab="Salience (log)", log="xy", xlim=c(5e-5,5e-2), ylim=c(0.001, 0.01), lty=c(1,2))
  points(x=crossPointX, y=crossPointY)
  text(x=crossPointX+1e-3, y=crossPointY, paste("O =",round(crossPointX,4)), pos=4)
dev.off()
