# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

source("Settings/training.R")
source("Datasets/datasets.R")
source("Cross-entropy/crossEntropyFunctions.R")

durationPties.nosmoothing = array(0, dim=c(2, 88, 2))
durationFrameCount = 0
for(analyzedFile in mutopiaFiles.training)
{
  cat("Filename =", analyzedFile, "\n")
  load(paste("Datasets/Mutopia/Processed/", analyzedFile, sep=""))
  for(rowNr in 1:88)
  {
    previousState = dataMatrix[rowNr, 1:(ncol(dataMatrix)-1)]
    currentState  = dataMatrix[rowNr, 2:ncol(dataMatrix)]
    for(kk in 0:1) for(ll in 0:1)
      durationPties.nosmoothing[ll+1, rowNr, kk+1] = durationPties.nosmoothing[ll+1, rowNr, kk+1] + length(which((previousState == kk) & (currentState == ll)))
  }
  durationFrameCount = durationFrameCount + ncol(dataMatrix) - 1
}

trainDurationModel = function(smoothingFactor=1)
{
  durationPties = (durationPties.nosmoothing + smoothingFactor) / (durationFrameCount + 4*smoothingFactor)  # +4 because of smoothing
  save(durationPties, file="TrainedParameters/durationModel.Rdata")
}

crossEntropyForSmoothingFactor = function(parameters)
{
  cat(" +", paste(rep("—", 160), collapse=""), "\n | Parameters = ", parameters[1], ", ", parameters[2], "\n", sep="")
  trainDurationModel(smoothingFactor=parameters[1])
  return(averageLikelihoodFunction(fudgeFactors=parameters[2], models="D"))
}

bestSmoothingFactor = optim(par=c(1,1), fn=crossEntropyForSmoothingFactor, lower=0)

# Use the default parameters
trainDurationModel(bestSmoothingFactor$par[1])

# load(file="Results/entropies.Rdata")
# load(file="Results/globalNotePties.Rdata")

# mutualInformations = rep(0, 88)
# for(ii in 1:88)
#   mutualInformations[ii] = sum(durationPties[,ii,] * log2(durationPties[,ii,] / outer(rowSums(durationPties[,ii,]), colSums(durationPties[,ii,]))))

# marginal1  = rep(0,2)
# marginal2  = rep(0,2)
# jointProbD = matrix(0, ncol=2, nrow=2)
# for(ii in 1:88)
# {
#   jointProbD = jointProbD + durationPties[,ii,]
#   marginal1  = marginal1 + rowSums(durationPties[,ii,])
#   marginal2  = marginal2 + colSums(durationPties[,ii,])
# }
# jointProbD = jointProbD / 88
# marginal1  = marginal1  / 88
# marginal2  = marginal2  / 88
# 
# mutualInformation = 0
# for(ii in 1:2) for(jj in 1:2)
#   mutualInformation = mutualInformation + jointProbD[ii,jj] * log2(jointProbD[ii,jj] / (marginal1[ii] * marginal2[jj]))
# 
# mutualInformations = rep(0, 88)
# for(ii in 1:88)
#   mutualInformations[ii] = sum(durationPties[,ii,] * log2(durationPties[,ii,] / outer(rowSums(durationPties[,ii,]), colSums(durationPties[,ii,]))))
# 
# meanNormalizedMutualInformation = sum(globalNotePties * (mutualInformations/entropies)) / sum(globalNotePties)
# meanMutualInformation = mean(mutualInformations)
# 
# if(plotResults)
# {
#   if(plotToPDF)
#   {
#     pdf("Plots/durationMI.pdf", width=10/2.54, height=6/2.54)
#     par(cex=0.5, lwd=0.5)
#   }
#   barplot(100 * mutualInformations/entropies, xlab="Piano key number", ylab="Normalized mutual information [%]", names.arg=paste(1:88))
#   abline(h=100*meanNormalizedMutualInformation, lty=2, col="dark red", lwd=1)
#   text(100, 100*meanNormalizedMutualInformation+3, paste(round(meanNormalizedMutualInformation*100,1), "%", sep=""), col="dark red")
#   if(plotToPDF)
#     dev.off()
# }
# 
