# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

source("Settings/training.R")
source("Datasets/datasets.R")
source("Cross-entropy/crossEntropyFunctions.R")

pianoPitches = c(10, 11, 12, rep(1:12, 7), 1)
octaveBegins = c(0, cumsum(c(3, rep(12, 7), 1))) + 1

allChromaEverMajor = matrix(0, 12, 0)
allChromaEverMinor = matrix(0, 12, 0)

if(exists("skippedFileNr")) {
  trainingRwcFiles = harmonyRwcFiles[-skippedFileNr]
} else {
  trainingRwcFiles = harmonyRwcFiles
}

# chordPties  = rep(0, 24)
# notePties   = rep(0, 88)
for(filename in trainingRwcFiles)
{
  cat(filename, "\n")
  load(paste("Datasets/RWC/Processed/", filename, ".txt.Rdata", sep=""))
  load(paste("Datasets/RWC/Chords/", filename, ".Rdata", sep=""))
  chords = chords + 1

  for(chordNr in 1:12)
  {
    if(any(chords == chordNr, na.rm=TRUE))
    {
      relevantFrames = which(chords == chordNr)
      chromaGram = matrix(0, 12, length(relevantFrames))
      chordData  = matrix(dataMatrix[, which(chords == chordNr)], ncol=length(relevantFrames))
      for(chromaNr in 1:12)
        chromaGram[chromaNr,] = colSums(matrix(chordData[which(pianoPitches == chromaNr), ], ncol=length(relevantFrames)))
      if(chordNr == 1)
        allChromaEverMajor = cbind(allChromaEverMajor, chromaGram)
      else
        allChromaEverMajor = cbind(allChromaEverMajor, chromaGram[c(chordNr:12,1:(chordNr-1)),])
    }
    if(any(chords == (chordNr + 12), na.rm=TRUE))
    {
      relevantFrames = which(chords == (chordNr + 12))
      chromaGram = matrix(0, 12, length(relevantFrames))
      chordData  = matrix(dataMatrix[, which(chords == (chordNr + 12))], ncol=length(relevantFrames))
      for(chromaNr in 1:12)
        chromaGram[chromaNr,] = colSums(matrix(chordData[which(pianoPitches == chromaNr), ], ncol=length(relevantFrames)))
      if(chordNr == 1)
        allChromaEverMinor = cbind(allChromaEverMinor, chromaGram)
      else
        allChromaEverMinor = cbind(allChromaEverMinor, chromaGram[c(chordNr:12,1:(chordNr-1)),])
    }
  }
#     # Marginal note distribution
#     for(noteNr in 1:88)
#       notePties[noteNr] = notePties[noteNr] + length(which(dataMatrix[noteNr,] == 1))
#     # Marginal chord distribution (for mutual information)
#     for(chordNr in 1:24)
#       chordPties[chordNr] = chordPties[chordNr] + length(which(chords == chordNr))
}

#   chordPties = chordPties / (ncol(allChromaEverMajor) + ncol(allChromaEverMinor))
#   notePties  = notePties  / (ncol(allChromaEverMajor) + ncol(allChromaEverMinor))

trainHarmonyModel = function(smoothingFactor=1)
{
  octaveCounts = c(8, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8)
  majorPties = (rowSums(allChromaEverMajor) + smoothingFactor*octaveCounts) / (2*smoothingFactor*octaveCounts + ncol(allChromaEverMajor))
  minorPties = (rowSums(allChromaEverMinor) + smoothingFactor*octaveCounts) / (2*smoothingFactor*octaveCounts + ncol(allChromaEverMinor))

  save(majorPties, file="TrainedParameters/harmonyModel.major.Rdata")
  save(minorPties, file="TrainedParameters/harmonyModel.minor.Rdata")
  return(list(majorPties, minorPties))
}

crossEntropyForSmoothingFactor = function(parameters)
{
  cat(" +", paste(rep("—", 160), collapse=""), "\n | Parameters = ", parameters[1], ", ", parameters[2], "\n", sep="")
  trainHarmonyModel(smoothingFactor=parameters[1])
  return(averageLikelihoodFunction(fudgeFactors=parameters[2], models="H"))
}

bestSmoothingFactor = optim(par=c(1,1), fn=crossEntropyForSmoothingFactor, method="L-BFGS-B", lower=0)

# Use the default parameters
noteCombinationLengthProbabilities = trainHarmonyModel(bestSmoothingFactor$par[1])

# ## Mutual information (doesn't work properly...!)
# noteProbabilities = matrix(NA, 24, 88)
# for(chordNr in 1:12)
# {
#   noteProbabilities[chordNr,]    = majorPties[(pianoPitches - chordNr) %% 12 + 1]
#   noteProbabilities[chordNr+12,] = minorPties[(pianoPitches - chordNr) %% 12 + 1]
# }
# 
# mutualInformations = colSums(noteProbabilities * log2(noteProbabilities / outer(chordPties, notePties)) + (1-noteProbabilities) * log2((1-noteProbabilities) / outer(chordPties, 1-notePties)))
