# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

source("Settings/training.R")
source("Datasets/datasets.R")
source("Cross-entropy/crossEntropyFunctions.R")

allData = matrix(NA, 88, 0)
for(filename in mutopiaFiles.training)
{
  filename = paste("Datasets/Mutopia/Processed/", filename, sep="")
  cat(filename, "\n")
  load(filename)
  allData = cbind(allData, dataMatrix)
}

noteCombinationLengthProbabilities.nonsmooth = rep(0, maxSimultaneousPitchNumber+1)
totalElements = rep(0, maxSimultaneousPitchNumber+1)

cat("  accumulativeNoteCombinationLengths...\n")
accumulativeNoteCombinationLengths = matrix(0, 88, ncol(allData))
for(pitchNr in 2:88)
  accumulativeNoteCombinationLengths[pitchNr, ] = accumulativeNoteCombinationLengths[pitchNr-1, ] + allData[pitchNr-1, ]

cat("  noteCombinationLengthProbabilities...\n")
for(combLength in 0:maxSimultaneousPitchNumber)  {
  indices = which(accumulativeNoteCombinationLengths == combLength)
  if(length(indices) > 0) {
    noteCombinationLengthProbabilities.nonsmooth[combLength+1] = noteCombinationLengthProbabilities.nonsmooth[combLength+1] + length(which(allData[indices] > 0))
    totalElements[combLength+1] = totalElements[combLength+1] + length(indices)
  }
}

trainPolyphonyModel = function(smoothingFactor)
{
  noteCombinationLengthProbabilities = (noteCombinationLengthProbabilities.nonsmooth + smoothingFactor) / (totalElements + 2*smoothingFactor)
  save(noteCombinationLengthProbabilities, file="TrainedParameters/polyphonyModel.Rdata")
}

crossEntropyForSmoothingFactor = function(parameters)
{
  cat("+", paste(rep("—", 160), collapse=""), "\n| Parameters = ", parameters[1], ", ", parameters[2], "\n", sep="")
  cat("Training probabilities...\n")
  trainPolyphonyModel(smoothingFactor=parameters[1])
  return(averageLikelihoodFunction(fudgeFactors=parameters[2], models="P"))
}
# 0.2156823
bestSmoothingFactor = optim(par=c(1,1), fn=crossEntropyForSmoothingFactor, lower=0)

# wyniki = sapply(exp(seq(0, 20, length.out=20))-1, function(x) crossEntropyForSmoothingFactor(c(x, 1)))
# plot(x=exp(seq(0, 20, length.out=20))-1, y=88*wyniki, log="x", type="l", xlab="Smoothing parameter", ylab="Average frame-wise cross-entropy")

# Save the usual results
trainPolyphonyModel(smoothingFactor=bestSmoothingFactor$par[1])

