# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

source("Settings/training.R")
source("Datasets/datasets.R")
source("Cross-entropy/crossEntropyFunctions.R")

maxDistance = 88
simpleVoiceProbs.nonsmooth = matrix(0, maxDistance+1, 2)

globalFrameCount = 0
fileNr = 1
for(analyzedFile in mutopiaFiles.training)
{
  cat("File ", fileNr, "/", length(mutopiaFiles.training), " (", analyzedFile, ")\n", sep="")
  load(paste("Datasets/Mutopia/Processed/", analyzedFile, sep=""))

  for(colNr in 1:(ncol(dataMatrix)-1))
  {
    activeNotes = which(dataMatrix[, colNr] > 0)
    if(length(activeNotes) > 0)
    {
      for(rowNr in 1:88)
      {
        closestNote = min(abs(rowNr - activeNotes)) + 1
        currentActivity = dataMatrix[rowNr, colNr+1] + 1
        simpleVoiceProbs.nonsmooth[closestNote, currentActivity] = simpleVoiceProbs.nonsmooth[closestNote, currentActivity] + 1
      }
    }
    else
    {
      for(rowNr in 1:88)
      {
        closestNote = maxDistance + 1
        currentActivity = dataMatrix[rowNr, colNr+1] + 1
        simpleVoiceProbs.nonsmooth[closestNote, currentActivity] = simpleVoiceProbs.nonsmooth[closestNote, currentActivity] + 1
      }
    }
  }
  fileNr = fileNr + 1
}

trainVoiceModel = function(smoothingFactor=1)
{
  simpleVoiceProbs = (simpleVoiceProbs.nonsmooth + smoothingFactor)
  simpleVoiceProbs = simpleVoiceProbs / rowSums(simpleVoiceProbs)
  save(simpleVoiceProbs, file="TrainedParameters/voiceModel.simple.Rdata")
}

crossEntropyForSmoothingFactor = function(parameters)
{
  cat("+", paste(rep("—", 160), collapse=""), "\n| Parameters = ", parameters[1], ", ", parameters[2], "\n", sep="")
  cat("Training probabilities...\n")
  trainVoiceModel(smoothingFactor=parameters[1])
  return(averageLikelihoodFunction(fudgeFactors=parameters[2], models="P"))
}

bestSmoothingFactor = optim(par=c(1,1), fn=crossEntropyForSmoothingFactor, lower=0)
trainVoiceModel(bestSmoothingFactor$par[1])

# # Obtain the mutual information
# simpleVoiceProbs = simpleVoiceProbs / sum(simpleVoiceProbs)
# mutualInformation = sum(simpleVoiceProbs * log2(simpleVoiceProbs / outer(rowSums(simpleVoiceProbs), colSums(simpleVoiceProbs))))

if(plotResults)
{
  pdf("Plots/voiceModel.pdf", width=imageWidth/2.54, height=6/2.54)
  par(cex=0.5, lwd=0.5)
  barplot(simpleVoiceProbs[1:37,2], pch=20, xlab="Distance to previous closest note", ylab="Activation probability", names.arg=0:36, log="y")
  load("TrainedParameters/priorOnProbability.Rdata")
  abline(h=priorOnProbability, col="dark red", lty=2, lwd=1)
  dev.off()
}

