# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

library('multicore')

source("Settings/audio.R")
source("Auxiliary/PRF.R")
source("Training/probabilityDefinitions.R")
load(file="TrainedParameters/onProbability.Rdata")
load(file="TrainedParameters/offProbability.Rdata")

# source("Datasets/datasets.R")
mutopiaFiles.testing = unlist(read.table("Datasets/Mutopia/mutopiaFiles.testing.txt", stringsAsFactors=FALSE), use.names=FALSE)

################################################################################
## Optimize threshold value for the entire testing data set

thresholdingResults.varyingThreshold = function(filename, thresholds, preselectionFunction=NULL, expFactor=NULL, maxFrameCount=NULL)
{
  processedFilename = gsub("/", "_", filename)
  processedFilename = gsub("\\.mid", "", processedFilename)
  load(paste("Datasets/Mutopia/Processed/", processedFilename, ".txt.Rdata", sep=""))
  processedFilename = gsub("\\.", "_", processedFilename)
  load(paste("Datasets/Mutopia/NMF/", processedFilename, ".Rdata", sep=""))

  if(ncol(audiodata) > ncol(dataMatrix))
    audiodata = audiodata[,1:ncol(dataMatrix)]
  if(ncol(dataMatrix) > ncol(audiodata))
    dataMatrix = dataMatrix[,1:ncol(audiodata)]

  if(!is.null(maxFrameCount) && (maxFrameCount < ncol(audiodata)))
  {
    audiodata  = audiodata[,1:maxFrameCount]
    dataMatrix = dataMatrix[,1:maxFrameCount]
  }

  if(!is.null(preselectionFunction))
  {
    if(!is.null(expFactor))
      audiodata = preselectionFunction(audiodata, expFactor=expFactor)
    else
      audiodata = preselectionFunction(audiodata)
  }

  resultingMatrix.threshold = matrix(NA, nrow=nrow(audiodata), ncol=ncol(audiodata))
  for(pitchNr in 1:88)
    resultingMatrix.threshold[pitchNr,] = ifelse(audiodata[pitchNr,] > thresholds[pitchNr], 1, 0)

  evaluationResults = c(PRF.onsets(resultingMatrix.threshold, dataMatrix))
  cat(evaluationResults[3], " ", sep="")
  return(evaluationResults)
}

thresholdingResults.givenThreshold = function(filename, threshold, preselectionFunction, expFactor=NULL, maxFrameCount=NULL)
  thresholdingResults.varyingThreshold(filename=filename, thresholds=rep(threshold, 88), preselectionFunction=preselectionFunction, expFactor=expFactor, maxFrameCount=maxFrameCount)

getSingleResults = function(dataset, threshold)
{
  cat(paste(rep("-", 120), collapse=""), "\n")
  cat("Threshold =", threshold, "\n\t")
  fmeasure = mean(matrix(unlist(mclapply(dataset, function(filename) thresholdingResults.givenThreshold(filename, threshold=threshold))), nrow=3)[3,])
  cat("\nF-measure =", fmeasure, "\n")
  return(fmeasure)
}

if(file.exists("Results/audio.T.bestThreshold.Rdata"))
{
  load(file="Results/audio.T.bestThreshold.Rdata")
} else
{
  bestThreshold = optimize(function(x) {
    getSingleResults(mutopiaFiles.testing, x)
  }, interval=c(0,1), maximum=TRUE)
  save(bestThreshold, file="Results/audio.T.bestThreshold.Rdata")
}

################################################################################
## Optimize exponential fudge factor for the pitch-dependent Bernoulli model

breakPos = seq(0, maxS, length.out=quantizationLevels+1)
binWidth = breakPos[2] - breakPos[1]
binMids  = seq(from=binWidth/2, by=binWidth, length.out=quantizationLevels)
realValues = binMids^(1/nonlinearityFactor)

pitchDependentBernoulli = function(exponentialFudgeFactor, pitchDependentSalience=FALSE, preselectionFunction=NULL, expFactor=NULL, maxFrameCount=maxFrameCount)
{
  if(pitchDependentSalience)
  {
    onProbabilities.fudged  = onProbabilities^exponentialFudgeFactor
    offProbabilities.fudged = offProbabilities^exponentialFudgeFactor
  }
  else
  {
    onProbability.fudged  = onProbability^exponentialFudgeFactor
    offProbability.fudged = offProbability^exponentialFudgeFactor
  }

  if(pitchDependentSalience)
    thresholds = sapply(1:88, function(pitchNr) realValues[which(onProbabilities.fudged[pitchNr,] * priorOnProbabilities[pitchNr] > offProbabilities.fudged[pitchNr,] * priorOffProbabilities[pitchNr])[1]])
  else
    thresholds = sapply(1:88, function(pitchNr) realValues[which(onProbability.fudged * priorOnProbabilities[pitchNr] > offProbability.fudged * priorOffProbabilities[pitchNr])[1]])

  # Get rid of the NAs at the low and high pitches:
  thresholds[which(is.na(thresholds[1:44]))] = max(thresholds[1:44], na.rm=TRUE)
  thresholds[which(is.na(thresholds[45:88]))+44] = max(thresholds[45:88], na.rm=TRUE)

  return(t(simplify2array(mclapply(mutopiaFiles.testing, function(filename) thresholdingResults.varyingThreshold(filename, thresholds=thresholds, preselectionFunction=preselectionFunction, expFactor=expFactor, maxFrameCount=maxFrameCount)))))
}

if(file.exists("Results/audio.PB.bestExponentialFudgeFactor.Rdata"))
{
  load(file="Results/audio.PB.bestExponentialFudgeFactor.Rdata")
} else
{
  bestExponentialFudgeFactor = optimize(function(x) {
    cat(paste(rep("-", 120), collapse=""), "\n")
    cat("Exponential factor =", x, "\n")
    meanFmeas = mean(pitchDependentBernoulli(x)[,3])
    cat("\nMean F-measure =", meanFmeas, "\n")
    return(meanFmeas)
  }, interval=c(0, 10), maximum=TRUE)
  save(bestExponentialFudgeFactor, file="Results/audio.PB.bestExponentialFudgeFactor.Rdata")
}

################################################################################
## Optimize exponential fudge factor for the pitch-dependent Bernoulli model w/ pitch-dependent salience model

load(file="TrainedParameters/onProbabilities.Rdata")
load(file="TrainedParameters/offProbabilities.Rdata")

if(file.exists("Results/audio.PPB.bestExponentialFudgeFactor.Rdata"))
{
  load(file="Results/audio.PPB.bestExponentialFudgeFactor.Rdata")
} else
{
  bestExponentialFudgeFactorPS = optimize(function(x) {
    cat(paste(rep("-", 120), collapse=""), "\n")
    cat("Exponential factor =", x, "\n")
    meanFmeas = mean(pitchDependentBernoulli(x, pitchDependentSalience=TRUE)[,3])
    cat("\nMean F-measure =", meanFmeas, "\n")
    return(meanFmeas)
  }, interval=c(0, 10), maximum=TRUE)
  save(bestExponentialFudgeFactorPS, file="Results/audio.PPB.bestExponentialFudgeFactor.Rdata")
}

################################################################################
## Obtain all results (no note preselection method)

evaluateSingleModel = function(evaluationFunction, trainSmoothingFactor=FALSE, modelName="")
{
  if(trainSmoothingFactor)
  {
    cat("Optimizing the exponential smoothing factor...\n")
    bestSmoothingFactor = optimize(f=function(expFactor) {
      cat("expFactor =", expFactor, "\n")
      meanFmeas = mean(evaluationFunction(expFactor=expFactor)[,3])
      cat("\n Mean F-measure = ", round(100*meanFmeas, 1), "%\n", sep="")
      return(meanFmeas)
    }, interval=c(0,1), maximum=TRUE)
    cat("Best smoothing factor =", bestSmoothingFactor$max, "\n")
    save(bestSmoothingFactor, file=paste("Results/audio.bestSmoothingFactor.", modelName, ".Rdata", sep=""))
    return(evaluationFunction(expFactor=bestSmoothingFactor$max))
  }
  else
    return(evaluationFunction())
}

modelNames = c("ML", "PB", "PML", "PPB", "T")
evaluateAllModels = function(preselectionFunction=NULL, trainSmoothingFactor=FALSE, maxFrameCount=NULL)
{
  modelEvaluation = array(NA, dim=c(length(modelNames), length(mutopiaFiles.testing), 3))
  dimnames(modelEvaluation) = list(modelNames)

  evaluationFunctions = c(
    function(expFactor=NULL) t(simplify2array(mclapply(
      mutopiaFiles.testing, function(filename) thresholdingResults.givenThreshold(filename,
                                                                                  threshold=realValues[which(onProbability > offProbability)[1]],
                                                                                  preselectionFunction=preselectionFunction,
                                                                                  expFactor=expFactor, maxFrameCount=maxFrameCount)))),
    function(expFactor=NULL) pitchDependentBernoulli(bestExponentialFudgeFactor$maximum,
                                                     preselectionFunction=preselectionFunction,
                                                     expFactor=expFactor,
                                                     maxFrameCount=maxFrameCount
                                                    ),
    function(expFactor=NULL) t(simplify2array(lapply(
      mutopiaFiles.testing, function(filename) thresholdingResults.varyingThreshold(filename,
                                                                                    thresholds=sapply(1:88, function(pitchNr) realValues[which(onProbabilities[pitchNr,] > offProbabilities[pitchNr,])[1]]),
                                                                                    preselectionFunction=preselectionFunction,
                                                                                    expFactor=expFactor, maxFrameCount=maxFrameCount)))),
    function(expFactor=NULL) pitchDependentBernoulli(bestExponentialFudgeFactorPS$maximum,
                                                     pitchDependentSalience=TRUE,
                                                     preselectionFunction=preselectionFunction,
                                                     expFactor=expFactor,
                                                     maxFrameCount=maxFrameCount
                                                    ),
    function(expFactor=NULL) t(simplify2array(mclapply(
      mutopiaFiles.testing, function(filename) thresholdingResults.givenThreshold(filename,
                                                                                  threshold=bestThreshold$maximum,
                                                                                  preselectionFunction=preselectionFunction,
                                                                                  expFactor=expFactor, maxFrameCount=maxFrameCount))))
  )
  names(evaluationFunctions) = modelNames

  for(modelName in modelNames)
  {
    cat("Obtaining results for ", modelName, "...\n", sep="")
    modelEvaluation[modelName,,] = evaluateSingleModel(evaluationFunctions[[modelName]], trainSmoothingFactor=trainSmoothingFactor, modelName=modelName)
  }
  cat("\n")

  return(modelEvaluation)
}

################################################################################
## Note preselection methods comparison

preselectionStandard = function(audiodata, expFactor=0.1, preselectedPeaks=6)
{
  S = audiodata^nonlinearityFactor
  source("Auxiliary/rightSmoothing.R", local=TRUE)  # will use the local expFactor
  for(timeSlice in 1:ncol(S))
  {
    highestPeaks = sort(smoothedS[, timeSlice], index.return=T, decreasing=T)$ix[1:preselectedPeaks]
    highestPeaks = highestPeaks[smoothedS[highestPeaks, timeSlice] > minimalSelectedSValue]
    audiodata[-highestPeaks,timeSlice] = 0
  }
  return(audiodata)
}

preselectionProbabilistic = function(audiodata, exponentialFudgeFactor=1, priorModel=0.5, preselectedPeaks=6, pitchDependentSalience=FALSE, expFactor=NULL) # expFactor is not used, but it is there for the sake of compatibility with preselectionStandard; removing it will cause execution errors
{
  S = ceiling(audiodata^nonlinearityFactor * quantizationLevels)
  S[S == 0] = 1   # bugfix: infinitesimal values are rounded to 1s, but 0s to 0s and 0s cause problems when used as indices

  if(pitchDependentSalience)
  {
    onProbability.fudged  = onProbabilities^exponentialFudgeFactor
    offProbability.fudged = offProbabilities^exponentialFudgeFactor
    normalizationVector = 1 / (onProbability.fudged + offProbability.fudged)  # it's actually a matrix
  }
  else
  {
    onProbability.fudged  = onProbability^exponentialFudgeFactor
    offProbability.fudged = offProbability^exponentialFudgeFactor
    normalizationVector = 1 / (onProbability.fudged + offProbability.fudged)
  }
  realOnProbability  = onProbability.fudged * normalizationVector
  realOffProbability = offProbability.fudged * normalizationVector

  priorOffModel = 1 - priorModel
  if(pitchDependentSalience)
    thresholds = sapply(1:88, function(pitchNr) realValues[which(realOnProbability[pitchNr,] * priorModel[pitchNr] > realOffProbability[pitchNr,] * priorOffModel[pitchNr])[1]])
  else
    thresholds = sapply(1:88, function(pitchNr) realValues[which(realOnProbability * priorModel[pitchNr] > realOffProbability * priorOffModel[pitchNr])[1]])

  thresholds[which(is.na(thresholds[1:44]))] = max(thresholds[1:44], na.rm=TRUE)
  thresholds[which(is.na(thresholds[45:88]))+44] = max(thresholds[45:88], na.rm=TRUE)

  if(pitchDependentSalience)
    probabilityMatrix = t(sapply(1:88, function(pitchNr) realOnProbability[pitchNr, S[pitchNr,]]))
  else
    probabilityMatrix = matrix(realOnProbability[S], nrow=88)

  for(timeSlice in 1:ncol(audiodata))
  {
    highestPeaks = sort(probabilityMatrix[, timeSlice], index.return=T, decreasing=T)$ix[1:preselectedPeaks]
    highestPeaks = highestPeaks[audiodata[highestPeaks, timeSlice] > thresholds[highestPeaks]]
    if(length(highestPeaks) > 0)
      audiodata[-highestPeaks,timeSlice] = 0
    else
      audiodata[,timeSlice] = 0
  }

  return(audiodata)
}

################################################################################
## Main code (doing all the computations)

maxFrameCount=320

preselectionFunctionNames = c("None", "Standard", "ML", "PB", "PML", "PPB")
if(file.exists("Results/audio.preselectionEvaluation.320.Rdata"))
{
  load(file="Results/audio.preselectionEvaluation.320.Rdata")
} else
{
  preselectionEvaluation = array(NA, dim=c(length(modelNames), length(mutopiaFiles.testing), length(preselectionFunctionNames), 3))
  dimnames(preselectionEvaluation) = list(modelNames, NULL, preselectionFunctionNames, c("Precission", "Recall", "F-measure"))

  cat(paste(rep("*", 120), collapse=""), "\n* Evaluating models for preselection type 'none'...\n", sep="")
  preselectionEvaluation[,,"None",]     = evaluateAllModels(maxFrameCount=maxFrameCount)
  cat(paste(rep("*", 120), collapse=""), "\n* Evaluating models for preselection type 'standard'...\n", sep="")
  preselectionEvaluation[,,"Standard",] = evaluateAllModels(preselectionFunction=preselectionStandard, trainSmoothingFactor=TRUE, maxFrameCount=maxFrameCount)
  cat(paste(rep("*", 120), collapse=""), "\n* Evaluating models for preselection type 'ML'...\n", sep="")
  preselectionEvaluation[,,"ML",]       = evaluateAllModels(preselectionFunction=preselectionProbabilistic, maxFrameCount=maxFrameCount)
  cat(paste(rep("*", 120), collapse=""), "\n* Evaluating models for preselection type 'PB'...\n", sep="")
  preselectionEvaluation[,,"PB",]       = evaluateAllModels(preselectionFunction=function(x) preselectionProbabilistic(x, exponentialFudgeFactor=bestExponentialFudgeFactor$maximum, priorModel=priorOnProbabilities), maxFrameCount=maxFrameCount)
  cat(paste(rep("*", 120), collapse=""), "\n* Evaluating models for preselection type 'PML'...\n", sep="")
  preselectionEvaluation[,,"PML",]      = evaluateAllModels(preselectionFunction=function(x) preselectionProbabilistic(x, pitchDependentSalience=TRUE), maxFrameCount=maxFrameCount)
  cat(paste(rep("*", 120), collapse=""), "\n* Evaluating models for preselection type 'PPB'...\n", sep="")
  preselectionEvaluation[,,"PPB",]      = evaluateAllModels(preselectionFunction=function(x) preselectionProbabilistic(x, pitchDependentSalience=TRUE, exponentialFudgeFactor=bestExponentialFudgeFactorPS$maximum, priorModel=priorOnProbabilities))
  save(preselectionEvaluation, file="Results/audio.preselectionEvaluation.320.Rdata")
}

################################################################################
## Plotting results

## Average PRF for no preselection
averageEvaluation = t(apply(preselectionEvaluation[,,"None",], c(1,3), mean))
averageEvaluation = averageEvaluation[,rev(modelNames)]  # it's better to plot them in the opposite order
pdf("Plots/audioModels.pdf", width=6/2.54, height=10/2.54)
par(cex=0.5, mar=c(5,4,2,2) + 0.1)
  plotColors = c("dark red", "dark green", "black")
  barplot(100*as.vector(t(averageEvaluation)), col=rep(plotColors, each=5), xlim=c(0,100), space=c(rep(0.2, ncol(averageEvaluation)), 1, rep(0.2, ncol(averageEvaluation)-1), 1, rep(0.2, ncol(averageEvaluation)-1)), names.arg=rep(rev(modelNames), 3), horiz=TRUE, xlab="Precision, recall and F-measure [%]", axes=FALSE, las=2)
  axis(1)
  lapply(1:3, function(ii) {
    for(modelNr in 1:ncol(averageEvaluation))
      text(y = c(0.7, 7.5, 14.3)[ii] + (modelNr-1) * 1.2, x = 100*averageEvaluation[ii,modelNr], paste(round(100*averageEvaluation[ii,modelNr], 1), "%", sep=""), pos=4, col=plotColors[ii], xpd=TRUE)
  })
dev.off()

## Average F-measure for different (preselection method, model) pairs
preselectionEvaluationMeans = apply(preselectionEvaluation, c(1,3,4), mean)[,,3]
preselectionEvaluationMeans = preselectionEvaluationMeans[,rev(preselectionFunctionNames)]  # it's better to plot them in the opposite order
pdf("Plots/preselectionResults.pdf", width=10/2.54, height=10/2.54)
par(cex=0.5, mar=c(5,8,2,2) + 0.1)
  modelCount = length(modelNames)
  preselectCount = length(preselectionFunctionNames)
  plotColors = c("dark red", "dark green", "dark blue", "dark orange", "black")
  barplot(100*as.vector(t(preselectionEvaluationMeans)), names.arg=outer(rev(preselectionFunctionNames), modelNames, function(x,y) paste(x, y, sep=" -> ")), xlim=c(0,80), las=2, col=rep(plotColors, each=preselectCount), xlab="F-measure [%]", horiz=TRUE, space=c(rep(0.2, preselectCount), rep(c(1, rep(0.2, preselectCount-1)), modelCount-1)), axes=F)
  axis(1)
  lapply(1:modelCount, function(modelNr) {
    for(preselectNr in 1:preselectCount)
      text(y = seq(from=0.5, by=8.04, length.out=modelCount)[modelNr] + (preselectNr-1) * 1.2, x = 100*preselectionEvaluationMeans[modelNr,preselectNr], paste(round(100*preselectionEvaluationMeans[modelNr,preselectNr], 1), "%", sep=""), pos=4, col=plotColors[modelNr], xpd=TRUE)
  })
dev.off()

bestSmoothingFactors = lapply(modelNames, function(modelName) {
  load(file=paste("Results/audio.bestSmoothingFactor.", modelName, ".Rdata", sep=""))
  return(bestSmoothingFactor)
})

