# -----------------------------------------------------------------
# 
#      MULTIPLE PITCH ANALYSIS WITH DYNAMIC BAYESIAN NETWORKS
# 
# This package contains the code for multiple pitch analysis using
# methods described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Dynamic Bayesian
# networks for symbolic polyphonic pitch modelling,” INRIA
# Technical Report RT-430, 2012,
# URL: http://hal.inria.fr/hal-00728771/en
# 
# This code has been written by Stanisław Raczyński at INRIA, France.
# Contact information: stanislaw.raczynski@inria.fr
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# -----------------------------------------------------------------

source("Auxiliary/PRF.R")
source("Datasets/datasets.R")

fileNrs = c(1,2,3,4,5,8,9,10,11,12,13,16,17,23,27,29,30,31,32,35,36,38,39)
acousticParValues = seq(0.5, 2.0, length.out=21)
bernoulliParValues = seq(0.6, 1, length.out=20)
modelStructures = c("D", "N", "P", "V", "HC", "HCV", "HCDVPN")

if(file.exists("Results/allFinalResults.Rdata")) {
  load("Results/allFinalResults.Rdata")
} else {
  results = array(NA, dim=c(length(modelStructures), length(fileNrs), length(acousticParValues), length(bernoulliParValues), 3))

  kk = 1
  for(fileNr in fileNrs)
  {
   cat("Loading data for ", fileNr, "...\n", sep="")
   load(paste("Datasets/Mutopia/Processed/", mutopiaFiles.testing[fileNr], sep=""))
   for(ll in seq_along(modelStructures))
   {
     cat(" · ", modelStructures[ll], "\n", sep="")
     for(ii in seq_along(acousticParValues)) for(jj in seq_along(bernoulliParValues))
     {
       load(paste("Results/Latest/", modelStructures[ll], "/resultingMatrix_", fileNr, "_", acousticParValues[ii], "_", bernoulliParValues[jj], ".Rdata", sep=""))
       results[ll,kk,ii,jj,] = PRF.onsets(resultingMatrix, dataMatrix[,1:ncol(resultingMatrix)], maxDist=10)
     }
   }
   kk = kk + 1
  }
  save(results, file="Results/allFinalResults.Rdata")
}

plotResultsAcoustic = function(fileName, ylim, textx, texty, xlines, ylines, xLabel)
{
  pdf(fileName, width=15/2.54, height=22/2.54)
    par(mfrow=c(4,2))
    for(ll in 1:8)
    {
      if(ll > 1) {
        bestAccIndices = apply(apply(results[ll-1,,,,3], c(2,3), mean), 1, which.max)
        resultPlot = sapply(1:3, function(prfNr) apply(results[ll-1,,,,], c(2,3,4), mean)[cbind(seq_len(dim(results)[3]), bestAccIndices, prfNr)])
      } else
        resultPlot = apply(results[3,,,20,], c(2,3), mean)
      if(ll %in% 1:6) xlab = "" else xlab = xLabel
      if(ll %in% c(1,3,5,7)) ylab="Precision, recall, F-measure [%]" else ylab=""
      if(ll == 1) par(mar=c(0,4,5,0)+0.1)
      if(ll == 2) par(mar=c(0,2,5,2)+0.1)
      if(ll == 3) par(mar=c(1.66,4,3.33,0)+0.1)
      if(ll == 4) par(mar=c(1.66,2,3.33,2)+0.1)
      if(ll == 5) par(mar=c(3.33,4,1.66,0)+0.1)
      if(ll == 6) par(mar=c(3.33,2,1.66,2)+0.1)
      if(ll == 7) par(mar=c(5,4,0,0)+0.1)
      if(ll == 8) par(mar=c(5,2,0,2)+0.1)
      matplot(x=acousticParValues, 100*resultPlot, type="l", lty=c(2,4,1), col=c("green", "red", "black"), ylab=ylab, xlab=xlab, ylim=ylim, axes=F)
      text(textx, texty, c("PB", modelStructures)[ll], pos=2, font=2)
      if(ll %in% c(1,3,5,7)) axis(2)
      if(ll %in% c(7,8)) axis(1)
      box()
      for(xline in xlines) abline(v=xline, lty=3, col="light gray")
      for(yline in ylines) abline(h=yline, lty=3, col="light gray")
    }
  dev.off()
}

plotResultsBernoulli = function(fileName, ylim, textx, texty, xlines, ylines, xLabel)
{
  bestAcoustic = which.max(apply(apply(apply(results[,,,,3], c(1,3,4), mean), c(2,3), sum), 1, max))
  pdf(fileName, width=15/2.54, height=22/2.54)
    par(mfrow=c(4,2))
    for(ll in 1:7)
    {
      resultPlot = apply(results[ll,,bestAcoustic,,], c(2,3), mean)
      if(ll %in% 1:5) xlab = "" else xlab = xLabel
      if(ll %in% c(1,2,4,6)) ylab="Precision, recall, F-measure [%]" else ylab=""
      if(ll == 1) par(mar=c(0,4,5,0)+0.1)
      if(ll == 2) frame() # par(mar=c(0,2,5,2)+0.1)
      if(ll == 2) par(mar=c(1.66,4,3.33,0)+0.1)
      if(ll == 3) par(mar=c(1.66,2,3.33,2)+0.1)
      if(ll == 4) par(mar=c(3.33,4,1.66,0)+0.1)
      if(ll == 5) par(mar=c(3.33,2,1.66,2)+0.1)
      if(ll == 6) par(mar=c(5,4,0,0)+0.1)
      if(ll == 7) par(mar=c(5,2,0,2)+0.1)
      matplot(x=bernoulliParValues, 100*resultPlot, type="l", lty=c(2,4,1), col=c("green", "red", "black"), ylab=ylab, xlab=xlab, ylim=ylim, axes=F)
      text(textx, texty, modelStructures[ll], pos=2, font=2)
      if(ll %in% c(1,2,4,6)) axis(2)
      if(ll %in% c(6,7)) axis(1)
      box()
      for(xline in xlines) abline(v=xline, lty=3, col="light gray")
      for(yline in ylines) abline(h=yline, lty=3, col="light gray")
    }
  dev.off()
}

plotResultsBernoulli("Plots/acoustic_bernoulli.pdf", ylim=c(67,87), textx=1, texty=86.5, ylines=seq(67.5, 85, by=2.5), xlines=seq(0.6, 1.0, by=0.1), xLabel="Bernoulli interpolation factor")
plotResultsAcoustic("Plots/acoustic_acoustic.pdf", ylim=c(20,87), textx=2, texty=23, ylines=seq(20, 85, by=10), xlines=seq(0.5, 2.0, by=0.5), xLabel="Acoustic weighting factor")

modelMeans = apply(results[,,8,,], c(1,3,4), mean)
bernoulliModel = 100*apply(results[,,8,20,], c(1,3), mean)[1,]
finalResults = 100*sapply(1:3, function(ii) modelMeans[cbind(1:7,apply(modelMeans[,,3], 1, which.max),ii)])
finalResults = rbind(bernoulliModel, finalResults)
plotIt = function(add=FALSE)
  barplot(as.vector(t(finalResults) - finalResults[8,]), names.arg=as.vector(rbind(rep(NA, 8), c("PB", modelStructures), rep(NA, 8))), ylab="P, R, F [%]", col=c("light gray", "light gray", "dark gray"), space=c(0,rep(c(0,0,0.6), 7), 0,0), density=c(25,25,-1), angle=c(45,-45,0), add=add)
pdf("Plots/modelFmeasure.pdf", width=8, height=5)
  plotIt(F)
  for(yy in seq(-10,5,by=1)) abline(h=yy, lty=3, col="light gray")
  plotIt(T)
  legend(22, -3, c("Precision", "Recall", "F-measure"), density=c(25,25,-1), angle=c(45,-45,0), col=c("light gray", "light gray", "dark gray"), bg="white")
dev.off()

lapply(1:3, function(prfNr) {
  pdf(paste("Plots/dupa", prfNr, ".pdf", sep=""))
    filled.contour(x=acousticParValues, y=bernoulliParValues, z=apply(results[3,,,,prfNr], c(2,3), mean), color.palette=function(n) gray.colors(n,1,0,2.2), nlevels=30, plot.axes={ 
      points(acousticParValues[apply(apply(results[3,,,,3], c(2,3), mean), 2, which.max)], bernoulliParValues, col="red", pch=19)
      axis(1); axis(2); box()
    })
  dev.off()
})

polyphonies = rep(NA, length(fileNrs))
ii = 1
for(fileNr in fileNrs)
{
  cat("Loading data for ", fileNr, "...\n", sep="")
  load(paste("Datasets/Mutopia/Processed/", mutopiaFiles.testing[fileNr], sep=""))
  polyphonies[ii] = mean(colSums(dataMatrix[,1:ifelse(ncol(dataMatrix) < 320, ncol(dataMatrix), 320)]))
  ii = ii + 1
}

individualResults = sapply(1:7, function(modelNr) results[modelNr,,8,,3][,which.max(apply(results[modelNr,,8,,3], 2, mean))])

result1 = 100*results[1,,8,20,3]
result2 = 100*individualResults[,1]

